/*
 * File: mux_driver.c
 *
 * Portions derived from rfcomm.c, original header as follows:
 *
 * Copyright (C) 2000, 2001  Axis Communications AB
 *
 * Author: Mats Friden <mats.friden@axis.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Exceptionally, Axis Communications AB grants discretionary and
 * conditional permissions for additional use of the text contained
 * in the company's release of the AXIS OpenBT Stack under the
 * provisions set forth hereunder.
 *
 * Provided that, if you use the AXIS OpenBT Stack with other files,
 * that do not implement functionality as specified in the Bluetooth
 * System specification, to produce an executable, this does not by
 * itself cause the resulting executable to be covered by the GNU
 * General Public License. Your use of that executable is in no way
 * restricted on account of using the AXIS OpenBT Stack code with it.
 *
 * This exception does not however invalidate any other reasons why
 * the executable file might be covered by the provisions of the GNU
 * General Public License.
 *
 */
/*
 * Copyright (C) 2002-2004  Motorola
 * Copyright (C) 2006 Harald Welte <laforge@openezx.org>
 * Copyright (C) 2007 Jim Rayner <jimr@beyondvoice.com>
 *
 *  07/28/2002  Initial version
 *  11/18/2002  Second version
 *  04/21/2004  Add GPRS PROC
 */

/* Note this code currently reuses the N_MOUSE ldisc for now */

#define TS0710DEBUG
#define PRINT_OUTPUT_PRINTK

#include <linux/version.h>

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
#include <linux/config.h>
#endif

#include <linux/module.h>
#include <linux/types.h>

#include <linux/kernel.h>
#include <linux/proc_fs.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/workqueue.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/spinlock.h>
#include <asm/termios.h> /* defined N_... ldiscs */
#include <linux/poll.h>

#include "ts0710_mux_ldisc.h"


#define TS0710MUX_MAJOR 250
#define NR_MUXS 16

				  /*#define TS0710MUX_TIME_OUT 30 *//* 300ms  */
#define TS0710MUX_TIME_OUT 250	/* 2500ms, for BP UART hardware flow control AP UART  */

#define TS0710MUX_IO_DLCI_FC_ON 0x54F2
#define TS0710MUX_IO_DLCI_FC_OFF 0x54F3
#define TS0710MUX_IO_FC_ON 0x54F4
#define TS0710MUX_IO_FC_OFF 0x54F5

#define TS0710MUX_SEND_BUF_OFFSET 10
#define TS0710MUX_SEND_BUF_SIZE (DEF_TS0710_MTU + TS0710MUX_SEND_BUF_OFFSET + 34)
#define TS0710MUX_RECV_BUF_SIZE TS0710MUX_SEND_BUF_SIZE

#define TS0710MUX_MAX_BUF_SIZE 256

#define TS0710MUX_SERIAL_BUF_SIZE (DEF_TS0710_MTU + TS0710_MAX_HDR_SIZE)
#define ADDRESS_FIELD_OFFSET 1
#define SEQ_FIELD_SIZE 0
#define SLIDE_BP_SEQ_OFFSET 1

/* To make things easy we map the tty index to the dlci. Therefore do not allow the tty index of 0 */
#define TS0710MUX_MINOR_START 1

#define TS0710MUX_MAX_TOTAL_FRAME_SIZE (DEF_TS0710_MTU + TS0710_MAX_HDR_SIZE + FLAG_SIZE)
#define TS0710MUX_MAX_CHARS_IN_BUF 65535
#define TS0710MUX_THROTTLE_THRESHOLD DEF_TS0710_MTU

#define TEST_PATTERN_SIZE 250

#ifndef UNUSED_PARAM
#define UNUSED_PARAM(v) (void)(v)
#endif

typedef struct {
	volatile __u8 buf[TS0710MUX_SEND_BUF_SIZE];
	volatile __u8 *frame;
	unsigned long flags;
	volatile __u16 length;
	volatile __u8 filled;
	volatile __u8 dummy;	/* Alignment to 4*n bytes */
} mux_send_struct;

/* Bit number in flags of mux_send_struct */
#define BUF_BUSY 0

struct mux_recv_packet_tag {
	__u8 *data;
	__u32 length;
	struct mux_recv_packet_tag *next;
};
typedef struct mux_recv_packet_tag mux_recv_packet;

struct mux_recv_struct_tag {
	__u8 data[TS0710MUX_RECV_BUF_SIZE];
	__u32 length;
	__u32 total;
	mux_recv_packet *mux_packet;
	struct mux_recv_struct_tag *next;
	int no_tty;
	volatile __u8 post_unthrottle;
};
typedef struct mux_recv_struct_tag mux_recv_struct;

#define RECV_RUNNING 0
static unsigned long mux_recv_flags = 0;

static mux_send_struct *mux_send_info[NR_MUXS];
static volatile __u8 mux_send_info_flags[NR_MUXS];
static volatile __u8 mux_send_info_idx = NR_MUXS;

static mux_recv_struct *mux_recv_info[NR_MUXS];
static volatile __u8 mux_recv_info_flags[NR_MUXS];
static mux_recv_struct *mux_recv_queue = NULL;

static struct tty_driver *mux_driver;

static struct tty_struct *COMM_FOR_MUX_TTY;

/* forward decls */
static int get_from_inbuf_list(unsigned char *buf, int dst_count);
static int mux_ldisc_write(const unsigned char *buf, int count);
static int mux_ldisc_chars_in_buffer(struct tty_struct *tty);

/* Line discipline data */
typedef struct {
	struct list_head list;
	int size;
	unsigned char *body; 
} buf_list_t;

struct link_ldisc_data {

  	struct list_head 	in_buf_list;
	spinlock_t		in_buf_lock;
        int in_buf_size;
}; 
struct link_ldisc_data *ldisc_data_ptr;

/* TODO check all these required since duplicating functionality of tty_io */
static struct work_struct send_tqueue;
static struct work_struct receive_tqueue;
static struct work_struct post_recv_tqueue;

static struct tty_struct *mux_table[NR_MUXS];

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
static struct termios *mux_termios[NR_MUXS];
static struct termios *mux_termios_locked[NR_MUXS];
#else
static struct ktermios *mux_termios[NR_MUXS];
static struct ktermios *mux_termios_locked[NR_MUXS];
#endif

static volatile short int mux_tty[NR_MUXS];

#ifdef min
#undef min
#define min(a,b)    ( (a)<(b) ? (a):(b) )
#endif

static int send_ua(ts0710_con * ts0710, __u8 dlci);
static int send_dm(ts0710_con * ts0710, __u8 dlci);
static int send_sabm(ts0710_con * ts0710, __u8 dlci);
static int send_disc(ts0710_con * ts0710, __u8 dlci);
static void queue_uih(mux_send_struct * send_info, __u16 len,
		      ts0710_con * ts0710, __u8 dlci);
static int send_pn_msg(ts0710_con * ts0710, __u8 prior, __u32 frame_size,
		       __u8 credit_flow, __u8 credits, __u8 dlci, __u8 cr);
static int send_nsc_msg(ts0710_con * ts0710, mcc_type cmd, __u8 cr);
static void set_uih_hdr(short_frame * uih_pkt, __u8 dlci, __u32 len, __u8 cr);

static __u32 crc_check(__u8 * data, __u32 length, __u8 check_sum);
static __u8 crc_calc(__u8 * data, __u32 length);
static void create_crctable(__u8 table[]);

static void mux_sched_send(void);

static __u8 crctable[256];

static ts0710_con ts0710_connection;

static int valid_dlci(__u8 dlci)
{
	if ((dlci < TS0710_MAX_CHN) && (dlci > 0))
		return 1;
	else
		return 0;
}


#ifdef PRINT_OUTPUT_PRINTK
#define TS0710_DEBUG(fmt, arg...) printk(KERN_INFO "MUX %s: " fmt " " , __func__, ## arg)
#define TS0710_LOG(fmt, arg...) printk(KERN_WARNING "MUX %s: " fmt " " , __func__, ## arg)

static void TS0710_DEBUGHEX(__u8 * buf, int len)
{
	static unsigned char tbuf[TS0710MUX_MAX_BUF_SIZE];

	int i;
	int c;

	if (len <= 0) {
		return;
	}

	c = 0;
	for (i = 0; (i < len) && (c < (TS0710MUX_MAX_BUF_SIZE - 3)); i++) {
		sprintf(&tbuf[c], "%02x ", buf[i]);
		c += 3;
	}
	tbuf[c] = 0;
	TS0710_DEBUG("%s\n", tbuf);
}

#else
#define TS0710_DEBUG(fmt, arg...)
#define TS0710_LOG(fmt, arg...)
#define TS0710_DEBUGHEX(buf, len)
#endif  /* PRINT_OUTPUT_PRINTK */

#ifdef TS0710LOG
static void TS0710_LOGSTR_FRAME(__u8 send, __u8 * data, int len)
{
	short_frame *short_pkt;
	long_frame *long_pkt;
	__u8 *uih_data_start;
	__u32 uih_len;
	__u8 dlci;
	int pos;

	if (len <= 0) {
		return;
	}

	pos = 0;
	if (send) {
		pos += sprintf(&g_tbuf[pos], "<");
		short_pkt = (short_frame *) (data);
	} else {
		pos += sprintf(&g_tbuf[pos], ">%d ", *(data));
		short_pkt = (short_frame *) (data + ADDRESS_FIELD_OFFSET);
	}

	dlci = short_pkt->h.addr.server_chn;
	switch (CLR_PF(short_pkt->h.control)) {
	case SABM:
		pos += sprintf(&g_tbuf[pos], "C SABM %d ::", dlci);
		break;
	case UA:
		pos += sprintf(&g_tbuf[pos], "C UA %d ::", dlci);
		break;
	case DM:
		pos += sprintf(&g_tbuf[pos], "C DM %d ::", dlci);
		break;
	case DISC:
		pos += sprintf(&g_tbuf[pos], "C DISC %d ::", dlci);
		break;
	case UIH:
		if (!dlci) {
			pos += sprintf(&g_tbuf[pos], "C MCC %d ::", dlci);
		} else {

			if ((short_pkt->h.length.ea) == 0) {
				long_pkt = (long_frame *) short_pkt;
				uih_len = GET_LONG_LENGTH(long_pkt->h.length);
				uih_data_start = long_pkt->h.data;
			} else {
				uih_len = short_pkt->h.length.len;
				uih_data_start = short_pkt->data;
			}
			pos += sprintf(&g_tbuf[pos], "I %d D %d ::", dlci, uih_len);

		}
		break;
	default:
		pos += sprintf(&g_tbuf[pos], "N!!! %d ::", dlci);
		break;
	}

	if (len > (sizeof(g_tbuf) - pos - 1)) {
		len = (sizeof(g_tbuf) - pos - 1);
	}

	memcpy(&g_tbuf[pos], data, len);
	pos += len;
	g_tbuf[pos] = 0;

#ifdef PRINT_OUTPUT_PRINTK
	/* 0x00 byte in the string pointed by g_tbuf may truncate the print result */
	TS0710_LOG("%s\n", g_tbuf);
#endif
}
#else
#define TS0710_LOGSTR_FRAME(send, data, len)
#endif


static int basic_write(__u8 * buf, int len)
{
	int res;
        int total_len = len + 2;

	TS0710_DEBUG("len %d\n",len);

	buf[0] = TS0710_BASIC_FLAG;
	buf[len + 1] = TS0710_BASIC_FLAG;

	TS0710_DEBUGHEX(buf, total_len);

	res = mux_ldisc_write(buf, total_len);
	if (res != total_len) {
		TS0710_DEBUG("Write Error!\n");
		return -1;
	}

	return total_len;
}

/* Functions for the crc-check and calculation */

#define CRC_VALID 0xcf

static __u32 crc_check(__u8 * data, __u32 length, __u8 check_sum)
{
	__u8 fcs = 0xff;

	while (length--) {
		fcs = crctable[fcs ^ *data++];
	}
	fcs = crctable[fcs ^ check_sum];
	if (fcs == (uint) 0xcf) {	/*CRC_VALID) */
		TS0710_DEBUG("CRC check OK\n");
		return 0;
	} else {
		TS0710_DEBUG("CRC check failed\n");
		return 1;
	}
}

/* Calculates the checksum according to the ts0710 specification */

static __u8 crc_calc(__u8 * data, __u32 length)
{
	__u8 fcs = 0xff;

	while (length--) {
		fcs = crctable[fcs ^ *data++];
	}

	return 0xff - fcs;
}

/* Calulates a reversed CRC table for the FCS check */
/* TODO - use the static table from 27.010 */
static void create_crctable(__u8 table[])
{
	int i, j;

	__u8 data;
	__u8 code_word = (__u8) 0xe0;
	__u8 sr = (__u8) 0;

	for (j = 0; j < 256; j++) {
		data = (__u8) j;

		for (i = 0; i < 8; i++) {
			if ((data & 0x1) ^ (sr & 0x1)) {
				sr >>= 1;
				sr ^= code_word;
			} else {
				sr >>= 1;
			}

			data >>= 1;
			sr &= 0xff;
		}

		table[j] = sr;
		sr = 0;
	}
}

static void ts0710_reset_dlci(__u8 j)
{
	if (j >= TS0710_MAX_CHN)
		return;

	ts0710_connection.dlci[j].state = DISCONNECTED;
	ts0710_connection.dlci[j].flow_control = 0;
	ts0710_connection.dlci[j].mtu = DEF_TS0710_MTU;
	ts0710_connection.dlci[j].initiated = 0;
	ts0710_connection.dlci[j].initiator = 0;
	init_waitqueue_head(&ts0710_connection.dlci[j].open_wait);
	init_waitqueue_head(&ts0710_connection.dlci[j].close_wait);
}

static void ts0710_reset_con(void)
{
	__u8 j;

	ts0710_connection.initiator = 0;
	ts0710_connection.mtu = DEF_TS0710_MTU + TS0710_MAX_HDR_SIZE;
	ts0710_connection.be_testing = 0;
	ts0710_connection.test_errs = 0;
	init_waitqueue_head(&ts0710_connection.test_wait);

	for (j = 0; j < TS0710_MAX_CHN; j++) {
		ts0710_reset_dlci(j);
	}
}

static void ts0710_init(void)
{
	create_crctable(crctable);

	ts0710_reset_con();
}

static void ts0710_upon_disconnect(void)
{
	ts0710_con *ts0710 = &ts0710_connection;
	__u8 j;

	for (j = 0; j < TS0710_MAX_CHN; j++) {
		ts0710->dlci[j].state = DISCONNECTED;
		wake_up_interruptible(&ts0710->dlci[j].open_wait);
		wake_up_interruptible(&ts0710->dlci[j].close_wait);
	}
	ts0710->be_testing = 0;
	wake_up_interruptible(&ts0710->test_wait);
	ts0710_reset_con();
}

/* Sending packet functions */

/* Creates a UA packet and puts it at the beginning of the pkt pointer */

static int send_ua(ts0710_con * ts0710, __u8 dlci)
{
	__u8 buf[sizeof(short_frame) + FCS_SIZE + FLAG_SIZE];
	short_frame *ua;

	TS0710_DEBUG("Creating UA packet to DLCI %d\n", dlci);

	ua = (short_frame *) (buf + 1);
	ua->h.addr.ea = 1;
	ua->h.addr.cr = ((~(ts0710->initiator)) & 0x1);
	ua->h.addr.server_chn = dlci;
	ua->h.control = SET_PF(UA);
	ua->h.length.ea = 1;
	ua->h.length.len = 0;
	ua->data[0] = crc_calc((__u8 *) ua, SHORT_CRC_CHECK);

	return basic_write(buf, sizeof(short_frame) + FCS_SIZE);
}

/* Creates a DM packet and puts it at the beginning of the pkt pointer */

static int send_dm(ts0710_con * ts0710, __u8 dlci)
{
	__u8 buf[sizeof(short_frame) + FCS_SIZE + FLAG_SIZE];
	short_frame *dm;

	TS0710_DEBUG("Creating DM packet to DLCI %d\n", dlci);

	dm = (short_frame *) (buf + 1);
	dm->h.addr.ea = 1;
	dm->h.addr.cr = ((~(ts0710->initiator)) & 0x1);
	dm->h.addr.server_chn = dlci;
	dm->h.control = SET_PF(DM);
	dm->h.length.ea = 1;
	dm->h.length.len = 0;
	dm->data[0] = crc_calc((__u8 *) dm, SHORT_CRC_CHECK);

	return basic_write(buf, sizeof(short_frame) + FCS_SIZE);
}

static int send_sabm(ts0710_con * ts0710, __u8 dlci)
{
	__u8 buf[sizeof(short_frame) + FCS_SIZE + FLAG_SIZE];
	short_frame *sabm;

	TS0710_DEBUG("Creating SABM packet to DLCI %d\n", dlci);

	sabm = (short_frame *) (buf + 1);
	sabm->h.addr.ea = 1;
	sabm->h.addr.cr = ((ts0710->initiator) & 0x1);
	sabm->h.addr.server_chn = dlci;
	sabm->h.control = SET_PF(SABM);
	sabm->h.length.ea = 1;
	sabm->h.length.len = 0;
	sabm->data[0] = crc_calc((__u8 *) sabm, SHORT_CRC_CHECK);

	return basic_write(buf, sizeof(short_frame) + FCS_SIZE);
}

static int send_disc(ts0710_con * ts0710, __u8 dlci)
{
	__u8 buf[sizeof(short_frame) + FCS_SIZE + FLAG_SIZE];
	short_frame *disc;

	TS0710_DEBUG("Creating DISC packet to DLCI %d\n", dlci);

	disc = (short_frame *) (buf + 1);
	disc->h.addr.ea = 1;
	disc->h.addr.cr = ((ts0710->initiator) & 0x1);
	disc->h.addr.server_chn = dlci;
	disc->h.control = SET_PF(DISC);
	disc->h.length.ea = 1;
	disc->h.length.len = 0;
	disc->data[0] = crc_calc((__u8 *) disc, SHORT_CRC_CHECK);

	return basic_write(buf, sizeof(short_frame) + FCS_SIZE);
}

static int send_closedown_msg(void)
{
	__u8 buf[10];

	cld_msg *uih_pkt = (cld_msg *) (buf + 1);

	uih_pkt->s_head.addr.ea = 1;
	uih_pkt->s_head.addr.cr = 0x1;
	uih_pkt->s_head.addr.server_chn = 0;

	uih_pkt->s_head.control = CLR_PF(UIH);

	uih_pkt->s_head.length.ea = 1;
	uih_pkt->s_head.length.len = 2;

        /* multiplexer control channel command */
	uih_pkt->mcc_s_head.type.ea = 1;
	uih_pkt->mcc_s_head.type.cr = 1;
	uih_pkt->mcc_s_head.type.type = CLD;

	uih_pkt->mcc_s_head.length.ea = 1;
	uih_pkt->mcc_s_head.length.len = 1;

	uih_pkt->fcs = crc_calc((__u8 *) uih_pkt, SHORT_CRC_CHECK);

	return basic_write(buf, 6);
}

static void queue_uih(mux_send_struct * send_info, __u16 len,
		      ts0710_con * ts0710, __u8 dlci)
{
	__u32 size;
	TS0710_DEBUG("Creating UIH packet (%d bytes, dlc %d)\n",len, dlci);

	if (len > SHORT_PAYLOAD_SIZE) {
		long_frame *l_pkt;

		size = sizeof(long_frame) + len + FCS_SIZE;
		l_pkt = (long_frame *) (send_info->frame - sizeof(long_frame));
		set_uih_hdr((void *)l_pkt, dlci, len, ts0710->initiator);
		l_pkt->data[len] = crc_calc((__u8 *) l_pkt, LONG_CRC_CHECK);
		send_info->frame = ((__u8 *) l_pkt) - 1;
	} else {
		short_frame *s_pkt;

		size = sizeof(short_frame) + len + FCS_SIZE;
		s_pkt =
		    (short_frame *) (send_info->frame - sizeof(short_frame));
		set_uih_hdr((void *)s_pkt, dlci, len, ts0710->initiator);
		s_pkt->data[len] = crc_calc((__u8 *) s_pkt, SHORT_CRC_CHECK);
		send_info->frame = ((__u8 *) s_pkt) - 1;
	}
	send_info->length = size;
}

/* Multiplexer command packets functions */

/* Turns on the ts0710 flow control */

static int ts0710_fcon_msg(ts0710_con * ts0710, __u8 cr)
{
	__u8 buf[30];
	mcc_short_frame *mcc_pkt;
	short_frame *uih_pkt;
	__u32 size;

	size = sizeof(short_frame) + sizeof(mcc_short_frame) + FCS_SIZE;
	uih_pkt = (short_frame *) (buf + 1);
	set_uih_hdr(uih_pkt, CTRL_CHAN, sizeof(mcc_short_frame),
		    ts0710->initiator);
	uih_pkt->data[sizeof(mcc_short_frame)] =
	    crc_calc((__u8 *) uih_pkt, SHORT_CRC_CHECK);
	mcc_pkt = (mcc_short_frame *) (uih_pkt->data);

	mcc_pkt->h.type.ea = EA;
	mcc_pkt->h.type.cr = cr;
	mcc_pkt->h.type.type = FCON;
	mcc_pkt->h.length.ea = EA;
	mcc_pkt->h.length.len = 0;

	return basic_write(buf, size);
}

/* Turns off the ts0710 flow control */

static int ts0710_fcoff_msg(ts0710_con * ts0710, __u8 cr)
{
	__u8 buf[30];
	mcc_short_frame *mcc_pkt;
	short_frame *uih_pkt;
	__u32 size;

	size = (sizeof(short_frame) + sizeof(mcc_short_frame) + FCS_SIZE);
	uih_pkt = (short_frame *) (buf + 1);
	set_uih_hdr(uih_pkt, CTRL_CHAN, sizeof(mcc_short_frame),
		    ts0710->initiator);
	uih_pkt->data[sizeof(mcc_short_frame)] =
	    crc_calc((__u8 *) uih_pkt, SHORT_CRC_CHECK);
	mcc_pkt = (mcc_short_frame *) (uih_pkt->data);

	mcc_pkt->h.type.ea = 1;
	mcc_pkt->h.type.cr = cr;
	mcc_pkt->h.type.type = FCOFF;
	mcc_pkt->h.length.ea = 1;
	mcc_pkt->h.length.len = 0;

	return basic_write(buf, size);
}

/* Sends an PN-messages and sets the not negotiable parameters to their
   default values in ts0710 */

static int send_pn_msg(ts0710_con * ts0710, __u8 prior, __u32 frame_size,
		       __u8 credit_flow, __u8 credits, __u8 dlci, __u8 cr)
{
	__u8 buf[30];
	pn_msg *pn_pkt;
	__u32 size;
	TS0710_DEBUG
	    ("DLCI 0x%02x, prior:0x%02x, frame_size:%d, credit_flow:%x, credits:%d, cr:%x\n",
	     dlci, prior, frame_size, credit_flow, credits, cr);

	size = sizeof(pn_msg);
	pn_pkt = (pn_msg *) (buf + 1);

	set_uih_hdr((void *)pn_pkt, CTRL_CHAN,
		    size - (sizeof(short_frame) + FCS_SIZE), ts0710->initiator);
	pn_pkt->fcs = crc_calc((__u8 *) pn_pkt, SHORT_CRC_CHECK);

	pn_pkt->mcc_s_head.type.ea = 1;
	pn_pkt->mcc_s_head.type.cr = cr;
	pn_pkt->mcc_s_head.type.type = PN;
	pn_pkt->mcc_s_head.length.ea = 1;
	pn_pkt->mcc_s_head.length.len = 8;

	pn_pkt->res1 = 0;
	pn_pkt->res2 = 0;
	pn_pkt->dlci = dlci;
	pn_pkt->frame_type = 0;
	pn_pkt->credit_flow = credit_flow;
	pn_pkt->prior = prior;
	pn_pkt->ack_timer = 0;
	SET_PN_MSG_FRAME_SIZE(pn_pkt, frame_size);
	pn_pkt->credits = credits;
	pn_pkt->max_nbrof_retrans = 0;

	return basic_write(buf, size);
}

/* Send a Not supported command - command, which needs 3 bytes */

static int send_nsc_msg(ts0710_con * ts0710, mcc_type cmd, __u8 cr)
{
	__u8 buf[30];
	nsc_msg *nsc_pkt;
	__u32 size;

	size = sizeof(nsc_msg);
	nsc_pkt = (nsc_msg *) (buf + 1);

	set_uih_hdr((void *)nsc_pkt, CTRL_CHAN,
		    sizeof(nsc_msg) - sizeof(short_frame) - FCS_SIZE,
		    ts0710->initiator);

	nsc_pkt->fcs = crc_calc((__u8 *) nsc_pkt, SHORT_CRC_CHECK);

	nsc_pkt->mcc_s_head.type.ea = 1;
	nsc_pkt->mcc_s_head.type.cr = cr;
	nsc_pkt->mcc_s_head.type.type = NSC;
	nsc_pkt->mcc_s_head.length.ea = 1;
	nsc_pkt->mcc_s_head.length.len = 1;

	nsc_pkt->command_type.ea = 1;
	nsc_pkt->command_type.cr = cmd.cr;
	nsc_pkt->command_type.type = cmd.type;

	return basic_write(buf, size);
}

static int ts0710_msc_msg(ts0710_con * ts0710, __u8 value, __u8 cr, __u8 dlci)
{
	__u8 buf[30];
	msc_msg *msc_pkt;
	__u32 size;

	size = sizeof(msc_msg);
	msc_pkt = (msc_msg *) (buf + 1);

	set_uih_hdr((void *)msc_pkt, CTRL_CHAN,
		    sizeof(msc_msg) - sizeof(short_frame) - FCS_SIZE,
		    ts0710->initiator);

	msc_pkt->fcs = crc_calc((__u8 *) msc_pkt, SHORT_CRC_CHECK);

	msc_pkt->mcc_s_head.type.ea = 1;
	msc_pkt->mcc_s_head.type.cr = cr;
	msc_pkt->mcc_s_head.type.type = MSC;
	msc_pkt->mcc_s_head.length.ea = 1;
	msc_pkt->mcc_s_head.length.len = 2;

	msc_pkt->dlci.ea = 1;
	msc_pkt->dlci.cr = 1;
	msc_pkt->dlci.server_chn = dlci;

	msc_pkt->v24_sigs = value;

	return basic_write(buf, size);
}

static int ts0710_test_msg(ts0710_con * ts0710, __u8 * test_pattern, __u32 len,
			   __u8 cr, __u8 * f_buf /*Frame buf */ )
{
	__u32 size;

	if (len > SHORT_PAYLOAD_SIZE) {
		long_frame *uih_pkt;
		mcc_long_frame *mcc_pkt;

		size =
		    (sizeof(long_frame) + sizeof(mcc_long_frame) + len +
		     FCS_SIZE);
		uih_pkt = (long_frame *) (f_buf + 1);

		set_uih_hdr((short_frame *) uih_pkt, CTRL_CHAN, len +
			    sizeof(mcc_long_frame), ts0710->initiator);
		uih_pkt->data[GET_LONG_LENGTH(uih_pkt->h.length)] =
		    crc_calc((__u8 *) uih_pkt, LONG_CRC_CHECK);
		mcc_pkt = (mcc_long_frame *) uih_pkt->data;

		mcc_pkt->h.type.ea = EA;
		/* cr tells whether it is a commmand (1) or a response (0) */
		mcc_pkt->h.type.cr = cr;
		mcc_pkt->h.type.type = TEST;
		SET_LONG_LENGTH(mcc_pkt->h.length, len);
		memcpy(mcc_pkt->value, test_pattern, len);
	} else if (len > (SHORT_PAYLOAD_SIZE - sizeof(mcc_short_frame))) {
		long_frame *uih_pkt;
		mcc_short_frame *mcc_pkt;

		/* Create long uih packet and short mcc packet */
		size =
		    (sizeof(long_frame) + sizeof(mcc_short_frame) + len +
		     FCS_SIZE);
		uih_pkt = (long_frame *) (f_buf + 1);

		set_uih_hdr((short_frame *) uih_pkt, CTRL_CHAN,
			    len + sizeof(mcc_short_frame), ts0710->initiator);
		uih_pkt->data[GET_LONG_LENGTH(uih_pkt->h.length)] =
		    crc_calc((__u8 *) uih_pkt, LONG_CRC_CHECK);
		mcc_pkt = (mcc_short_frame *) uih_pkt->data;

		mcc_pkt->h.type.ea = EA;
		mcc_pkt->h.type.cr = cr;
		mcc_pkt->h.type.type = TEST;
		mcc_pkt->h.length.ea = EA;
		mcc_pkt->h.length.len = len;
		memcpy(mcc_pkt->value, test_pattern, len);
	} else {
		short_frame *uih_pkt;
		mcc_short_frame *mcc_pkt;

		size =
		    (sizeof(short_frame) + sizeof(mcc_short_frame) + len +
		     FCS_SIZE);
		uih_pkt = (short_frame *) (f_buf + 1);

		set_uih_hdr((void *)uih_pkt, CTRL_CHAN, len
			    + sizeof(mcc_short_frame), ts0710->initiator);
		uih_pkt->data[uih_pkt->h.length.len] =
		    crc_calc((__u8 *) uih_pkt, SHORT_CRC_CHECK);
		mcc_pkt = (mcc_short_frame *) uih_pkt->data;

		mcc_pkt->h.type.ea = EA;
		mcc_pkt->h.type.cr = cr;
		mcc_pkt->h.type.type = TEST;
		mcc_pkt->h.length.ea = EA;
		mcc_pkt->h.length.len = len;
		memcpy(mcc_pkt->value, test_pattern, len);

	}
	return basic_write(f_buf, size);
}

static void set_uih_hdr(short_frame * uih_pkt, __u8 dlci, __u32 len, __u8 cr)
{
	uih_pkt->h.addr.ea = 1;
	uih_pkt->h.addr.cr = cr;
	uih_pkt->h.addr.server_chn = dlci;
	uih_pkt->h.control = CLR_PF(UIH);

	if (len > SHORT_PAYLOAD_SIZE) {
		SET_LONG_LENGTH(((long_frame *) uih_pkt)->h.length, len);
	} else {
		uih_pkt->h.length.ea = 1;
		uih_pkt->h.length.len = len;
	}
}

/* Parses a multiplexer control channel packet */

void process_mcc(__u8 * data, __u32 len, ts0710_con * ts0710, int longpkt)
{
	__u8 *tbuf = NULL;
	mcc_short_frame *mcc_short_pkt;
	int j;

	if (longpkt) {
		mcc_short_pkt =
		    (mcc_short_frame *) (((long_frame *) data)->data);
	} else {
		mcc_short_pkt =
		    (mcc_short_frame *) (((short_frame *) data)->data);
	}

	switch (mcc_short_pkt->h.type.type) {
	case TEST:
		if (mcc_short_pkt->h.type.cr == MCC_RSP) {
			TS0710_DEBUG("Received test command response\n");

			if (ts0710->be_testing) {
				if ((mcc_short_pkt->h.length.ea) == 0) {
					mcc_long_frame *mcc_long_pkt;
					mcc_long_pkt =
					    (mcc_long_frame *) mcc_short_pkt;
					if (GET_LONG_LENGTH
					    (mcc_long_pkt->h.length) !=
					    TEST_PATTERN_SIZE) {
						ts0710->test_errs =
						    TEST_PATTERN_SIZE;
						TS0710_DEBUG
						    ("Err: received test pattern is %d bytes long, not expected %d\n",
						     GET_LONG_LENGTH
						     (mcc_long_pkt->h.length),
						     TEST_PATTERN_SIZE);
					} else {
						ts0710->test_errs = 0;
						for (j = 0;
						     j < TEST_PATTERN_SIZE;
						     j++) {
							if (mcc_long_pkt->
							    value[j] !=
							    (j & 0xFF)) {
								(ts0710->
								 test_errs)++;
							}
						}
					}

				} else {

#if TEST_PATTERN_SIZE < 128
					if (mcc_short_pkt->h.length.len !=
					    TEST_PATTERN_SIZE) {
#endif

						ts0710->test_errs =
						    TEST_PATTERN_SIZE;
						TS0710_DEBUG
						    ("Err: received test pattern is %d bytes long, not expected %d\n",
						     mcc_short_pkt->h.length.
						     len, TEST_PATTERN_SIZE);

#if TEST_PATTERN_SIZE < 128
					} else {
						ts0710->test_errs = 0;
						for (j = 0;
						     j < TEST_PATTERN_SIZE;
						     j++) {
							if (mcc_short_pkt->
							    value[j] !=
							    (j & 0xFF)) {
								(ts0710->
								 test_errs)++;
							}
						}
					}
#endif

				}

				ts0710->be_testing = 0;	/* Clear the flag */
				wake_up_interruptible(&ts0710->test_wait);
			} else {
				TS0710_DEBUG
				    ("Err: shouldn't or late to get test cmd response\n");
			}
		} else {
			tbuf = (__u8 *) kmalloc(len + 32, GFP_ATOMIC);
			if (!tbuf) {
				break;
			}

			if ((mcc_short_pkt->h.length.ea) == 0) {
				mcc_long_frame *mcc_long_pkt;
				mcc_long_pkt = (mcc_long_frame *) mcc_short_pkt;
				ts0710_test_msg(ts0710, mcc_long_pkt->value,
						GET_LONG_LENGTH(mcc_long_pkt->h.
								length),
						MCC_RSP, tbuf);
			} else {
				ts0710_test_msg(ts0710, mcc_short_pkt->value,
						mcc_short_pkt->h.length.len,
						MCC_RSP, tbuf);
			}

			kfree(tbuf);
		}
		break;

	case FCON:		/*Flow control on command */
		TS0710_DEBUG
		    ("Received Flow control(all channels) on command\n");
		if (mcc_short_pkt->h.type.cr == MCC_CMD) {
			ts0710->dlci[0].state = CONNECTED;
			ts0710_fcon_msg(ts0710, MCC_RSP);
			mux_sched_send();
		}
		break;

	case FCOFF:		/*Flow control off command */
		TS0710_DEBUG
		    ("Received Flow control(all channels) off command\n");
		if (mcc_short_pkt->h.type.cr == MCC_CMD) {
			for (j = 0; j < TS0710_MAX_CHN; j++) {
				ts0710->dlci[j].state = FLOW_STOPPED;
			}
			ts0710_fcoff_msg(ts0710, MCC_RSP);
		}
		break;

	case MSC:		/*Modem status command */
		{
			__u8 dlci;
			__u8 v24_sigs;

			dlci = (mcc_short_pkt->value[0]) >> 2;
			v24_sigs = mcc_short_pkt->value[1];

			if ((ts0710->dlci[dlci].state != CONNECTED)
			    && (ts0710->dlci[dlci].state != FLOW_STOPPED)) {
				send_dm(ts0710, dlci);
				break;
			}
			if (mcc_short_pkt->h.type.cr == MCC_CMD) {
				TS0710_DEBUG("Received Modem status command\n");
				if (v24_sigs & 2) {
					if (ts0710->dlci[dlci].state ==
					    CONNECTED) {
						TS0710_LOG
						    ("Received Flow off on dlci %d\n",
						     dlci);
						ts0710->dlci[dlci].state =
						    FLOW_STOPPED;
					}
				} else {
					if (ts0710->dlci[dlci].state ==
					    FLOW_STOPPED) {
						ts0710->dlci[dlci].state =
						    CONNECTED;
						TS0710_LOG
						    ("Received Flow on on dlci %d\n",
						     dlci);
						mux_sched_send();
					}
				}

				ts0710_msc_msg(ts0710, v24_sigs, MCC_RSP, dlci);
			} else {
				TS0710_DEBUG
				    ("Received Modem status response\n");

				if (v24_sigs & 2) {
					TS0710_DEBUG("Flow stop accepted\n");
				}
			}
			break;
		}
	case PN:		/*DLC parameter negotiation */
		{
			__u8 dlci;
			__u16 frame_size;
			pn_msg *pn_pkt;

			pn_pkt = (pn_msg *) data;
			dlci = pn_pkt->dlci;
			frame_size = GET_PN_MSG_FRAME_SIZE(pn_pkt);
			TS0710_DEBUG
			    ("Received DLC parameter negotiation, PN\n");
			if (pn_pkt->mcc_s_head.type.cr == MCC_CMD) {
				TS0710_DEBUG("received PN command with:\n");
				TS0710_DEBUG("frame size:%d\n", frame_size);

				frame_size =
				    min(frame_size, ts0710->dlci[dlci].mtu);
				send_pn_msg(ts0710, pn_pkt->prior, frame_size,
					    0, 0, dlci, MCC_RSP);
				ts0710->dlci[dlci].mtu = frame_size;
				TS0710_DEBUG("mtu set to %d\n",
					     ts0710->dlci[dlci].mtu);
			} else {
				TS0710_DEBUG("frame size:%d\n", frame_size);

				frame_size =
				    min(frame_size, ts0710->dlci[dlci].mtu);
				ts0710->dlci[dlci].mtu = frame_size;

				TS0710_DEBUG
				    ("mtu set on dlci:%d to %d\n",
				     dlci, ts0710->dlci[dlci].mtu);

				if (ts0710->dlci[dlci].state == NEGOTIATING) {
					ts0710->dlci[dlci].state = CONNECTING;
					wake_up_interruptible(&ts0710->
							      dlci[dlci].
							      open_wait);
				}
			}
			break;
		}

	case NSC:		/*Non supported command resonse */
		TS0710_LOG("received non supported command response\n");
		break;

	default:		/*Non supported command received */
		TS0710_LOG("received a non supported command\n");
		send_nsc_msg(ts0710, mcc_short_pkt->h.type, MCC_RSP);
		break;
	}
}

static mux_recv_packet *get_mux_recv_packet(__u32 size)
{
	mux_recv_packet *recv_packet;

	recv_packet =
	    (mux_recv_packet *) kmalloc(sizeof(mux_recv_packet), GFP_ATOMIC);
	TS0710_DEBUG("recv_packet %d\n",(int) recv_packet);
	if (!recv_packet) {
		return 0;
	}

	recv_packet->data = (__u8 *) kmalloc(size, GFP_ATOMIC);
	if (!(recv_packet->data)) {
		kfree(recv_packet);
		return 0;
	}
	recv_packet->length = 0;
	recv_packet->next = 0;
	return recv_packet;
}

static void free_mux_recv_packet(mux_recv_packet * recv_packet)
{
	TS0710_DEBUG("recv_packet %d\n",(int) recv_packet);
	if (!recv_packet) {
		return;
	}

	if (recv_packet->data) {
		kfree(recv_packet->data);
	}
	kfree(recv_packet);
}

static void free_mux_recv_struct(mux_recv_struct * recv_info)
{
	mux_recv_packet *recv_packet1, *recv_packet2;

	if (!recv_info) {
		return;
	}

	recv_packet1 = recv_info->mux_packet;
	while (recv_packet1) {
		recv_packet2 = recv_packet1->next;
		free_mux_recv_packet(recv_packet1);
		recv_packet1 = recv_packet2;
	}

	kfree(recv_info);
}

static inline void add_post_recv_queue(mux_recv_struct ** head,
				       mux_recv_struct * new_item)
{
	new_item->next = *head;
	*head = new_item;
}

static void ts0710_flow_on(__u8 dlci, ts0710_con * ts0710)
{
	int i;

	if ((ts0710->dlci[0].state != CONNECTED)
	    && (ts0710->dlci[0].state != FLOW_STOPPED)) {
		return;
	} else if ((ts0710->dlci[dlci].state != CONNECTED)
		   && (ts0710->dlci[dlci].state != FLOW_STOPPED)) {
		return;
	}

	if (!(ts0710->dlci[dlci].flow_control)) {
		return;
	}

	for (i = 0; i < 3; i++) {
		if (ts0710_msc_msg(ts0710, EA | RTC | RTR | DV, MCC_CMD, dlci) <
		    0) {
			continue;
		} else {
			TS0710_LOG("send Flow on on dlci %d\n", dlci);
			ts0710->dlci[dlci].flow_control = 0;
			break;
		}
	}
}

static void ts0710_flow_off(struct tty_struct *tty, __u8 dlci,
			    ts0710_con * ts0710)
{
	int i;

	if (test_and_set_bit(TTY_THROTTLED, &tty->flags)) {
		return;
	}

	if ((ts0710->dlci[0].state != CONNECTED)
	    && (ts0710->dlci[0].state != FLOW_STOPPED)) {
		return;
	} else if ((ts0710->dlci[dlci].state != CONNECTED)
		   && (ts0710->dlci[dlci].state != FLOW_STOPPED)) {
		return;
	}

	if (ts0710->dlci[dlci].flow_control) {
		return;
	}

	for (i = 0; i < 3; i++) {
		if (ts0710_msc_msg
		    (ts0710, EA | FC | RTC | RTR | DV, MCC_CMD, dlci) < 0) {
			continue;
		} else {
			TS0710_LOG("send Flow off on dlci %d\n", dlci);
			ts0710->dlci[dlci].flow_control = 1;
			break;
		}
	}
}

static int ts0710_recv_data(ts0710_con * ts0710, char *data, int len)
{
	long_frame *long_pkt;
	__u8 *uih_data_start;
	__u32 uih_len;
	__u8 be_connecting;
#ifdef TS0710DEBUG
	unsigned long t;
#endif

	short_frame *short_pkt = (short_frame *) data;
	__u8 dlci = short_pkt->h.addr.server_chn;

        TS0710_DEBUG("dlc %d\n",dlci);

	switch (CLR_PF(short_pkt->h.control)) {
	case SABM:
		TS0710_DEBUG("SABM-packet received\n");
		if (!dlci) {
			ts0710->dlci[0].state = CONNECTED;

			/* sending UA back */
			send_ua(ts0710, dlci);
			wake_up_interruptible(&ts0710->dlci[0].open_wait);

		} else if (valid_dlci(dlci)) {

			TS0710_DEBUG("Incoming connect on channel %d\n", dlci);
			/* sending UA back */
			send_ua(ts0710, dlci);

			ts0710->dlci[dlci].state = CONNECTED;
			wake_up_interruptible(&ts0710->dlci[dlci].open_wait);

		} else {
			TS0710_DEBUG("invalid dlci %d, sending DM\n", dlci);
			send_dm(ts0710, dlci);
		}
		break;

	case UA:
		TS0710_DEBUG("UA packet received\n");
		if (!dlci) {
			if (ts0710->dlci[0].state == CONNECTING) {
				ts0710->dlci[0].state = CONNECTED;
				wake_up_interruptible(&ts0710->dlci[0].
						      open_wait);
			} else if (ts0710->dlci[0].state == DISCONNECTING) {
				ts0710_upon_disconnect();
			} else {
				TS0710_DEBUG
				    (" Something wrong receiving UA packet\n");
			}
		} else if (valid_dlci(dlci)) {
			TS0710_DEBUG("Incoming UA on channel %d\n", dlci);

			if (ts0710->dlci[dlci].state == CONNECTING) {
				ts0710->dlci[dlci].state = CONNECTED;
				wake_up_interruptible(&ts0710->dlci[dlci].
						      open_wait);
			} else if (ts0710->dlci[dlci].state == DISCONNECTING) {
				ts0710->dlci[dlci].state = DISCONNECTED;
				wake_up_interruptible(&ts0710->dlci[dlci].
						      open_wait);
				wake_up_interruptible(&ts0710->dlci[dlci].
						      close_wait);
				ts0710_reset_dlci(dlci);
			} else {
				TS0710_DEBUG
				    (" Something wrong receiving UA packet\n");
			}
		} else {
			TS0710_DEBUG("invalid dlci %d\n", dlci);
		}
		break;

	case DM:
		TS0710_DEBUG("DM packet received\n");
		if (!dlci) {
			if (ts0710->dlci[0].state == CONNECTING) {
				be_connecting = 1;
			} else {
				be_connecting = 0;
			}
			ts0710_upon_disconnect();
			if (be_connecting) {
				ts0710->dlci[0].state = REJECTED;
			}
		} else if (valid_dlci(dlci)) {
			TS0710_DEBUG("Incoming DM on channel %d\n", dlci);

			if (ts0710->dlci[dlci].state == CONNECTING) {
				ts0710->dlci[dlci].state = REJECTED;
			} else {
				ts0710->dlci[dlci].state = DISCONNECTED;
			}
			wake_up_interruptible(&ts0710->dlci[dlci].open_wait);
			wake_up_interruptible(&ts0710->dlci[dlci].close_wait);
			ts0710_reset_dlci(dlci);
		} else {
			TS0710_DEBUG("invalid dlci %d\n", dlci);
		}
		break;

	case DISC:
		TS0710_DEBUG("DISC packet received\n");
		if (!dlci) {
			/* sending UA back */
			send_ua(ts0710, dlci);
			ts0710_upon_disconnect();
		} else if (valid_dlci(dlci)) {
			TS0710_DEBUG("Incoming DISC on channel %d\n", dlci);

			send_ua(ts0710, dlci);
			TS0710_DEBUG("DISC, sending back UA\n");

			ts0710->dlci[dlci].state = DISCONNECTED;
			wake_up_interruptible(&ts0710->dlci[dlci].open_wait);
			wake_up_interruptible(&ts0710->dlci[dlci].close_wait);
			ts0710_reset_dlci(dlci);
		} else {
			TS0710_DEBUG("invalid dlci %d\n", dlci);
		}
		break;

	case UIH:
		TS0710_DEBUG("UIH packet received\n");
		if ((dlci >= TS0710_MAX_CHN)) {
			TS0710_DEBUG("invalid dlci %d\n", dlci);
			send_dm(ts0710, dlci);
			break;
		}

		if (GET_PF(short_pkt->h.control)) {
			TS0710_LOG
			    ("Error UIH packet with P/F set, discard it!\n");
			break;
		}

		if ((ts0710->dlci[dlci].state != CONNECTED)
		    && (ts0710->dlci[dlci].state != FLOW_STOPPED)) {
			TS0710_LOG
			    ("Error DLCI %d not connected, discard it!\n", dlci);
			send_dm(ts0710, dlci);
			break;
		}

		if ((short_pkt->h.length.ea) == 0) {
			TS0710_DEBUG("Long UIH packet received\n");
			long_pkt = (long_frame *) data;
			uih_len = GET_LONG_LENGTH(long_pkt->h.length);
			uih_data_start = long_pkt->h.data;
			TS0710_DEBUG("long packet length %d\n", uih_len);
		} else {
			TS0710_DEBUG("Short UIH pkt received\n");
			uih_len = short_pkt->h.length.len;
			uih_data_start = short_pkt->data;
		}

		if (dlci == 0) {
			TS0710_DEBUG("UIH on serv_channel 0\n");
			process_mcc(data, len, ts0710,
				    !(short_pkt->h.length.ea));
		} else if (valid_dlci(dlci)) {
			/* do tty dispatch */
			__u8 tag;
			__u8 tty_idx;
			struct tty_struct *tty;
			__u8 queue_data;
			__u8 post_recv;
			__u8 flow_control;
			mux_recv_struct *recv_info;
			int recv_room;
			mux_recv_packet *recv_packet, *recv_packet2;

			TS0710_DEBUG("UIH on channel %d\n", dlci);

			if (uih_len > ts0710->dlci[dlci].mtu) {
				TS0710_DEBUG
				    ("Error:  DLCI:%d, uih_len:%d is bigger than mtu:%d, discard data!\n",
				     dlci, uih_len, ts0710->dlci[dlci].mtu);
				break;
			}

			tag = *uih_data_start;
			uih_data_start++;
			uih_len--;

			if (!uih_len) {
				break;
			}

			tty_idx = dlci;

			tty = mux_table[tty_idx];
			if ((!mux_tty[tty_idx]) || (!tty)) {
				TS0710_DEBUG("Port /dev/mux%d not open, discard it! \n",tty_idx);
				TS0710_DEBUG("mux_tty[%d] = %d, tty is at %i\n", tty_idx, mux_tty[tty_idx], (int) tty);
			} else {	/* Begin processing received data */
				if ((!mux_recv_info_flags[tty_idx])
				    || (!mux_recv_info[tty_idx])) {
					TS0710_DEBUG("Error: No mux_recv_info, discard it! /dev/mux%d\n",tty_idx);
					break;
				}

				recv_info = mux_recv_info[tty_idx];
				if (recv_info->total > 8192) {
					TS0710_DEBUG
					    ("discard data for tty_idx:%d, recv_info->total > 8192 \n",tty_idx);
					break;
				}

				queue_data = 0;
				post_recv = 0;
				flow_control = 0;
				recv_room = 65535;
				if (tty->receive_room)
					recv_room = tty->receive_room;

				if (test_bit(TTY_THROTTLED, &tty->flags)) {
					queue_data = 1;
				} else {
					if ((recv_room -
					     (uih_len + recv_info->total)) <
					    ts0710->dlci[dlci].mtu) {
						flow_control = 1;
					}
				}

				if (!queue_data) {
					/* Put received data into read buffer of tty */
					TS0710_DEBUG
					    ("Put received data into read buffer of /dev/mux%d\n",tty_idx);

#ifdef TS0710DEBUG
					t = jiffies;
#endif

					(tty->ldisc.receive_buf) (tty,
								  uih_data_start,
								  NULL,
								  uih_len);

#ifdef TS0710DEBUG
					TS0710_DEBUG
					    ("tty->ldisc.receive_buf (took ticks: %lu)\n",
					     (jiffies - t));
#endif

				} else {	/* Queue data */

					TS0710_DEBUG
					    ("Put received data into recv queue of /dev/mux%d\n",
					     tty_idx);
					if (recv_info->total) {
						/* recv_info is already linked into mux_recv_queue */

						recv_packet =
						    get_mux_recv_packet
						    (uih_len);
						if (!recv_packet) {
							TS0710_DEBUG("no memory\n");
							break;
						}

						memcpy(recv_packet->data,
						       uih_data_start, uih_len);
						recv_packet->length = uih_len;
						recv_info->total += uih_len;
						recv_packet->next = NULL;

						if (!(recv_info->mux_packet)) {
							recv_info->mux_packet =
							    recv_packet;
						} else {
							recv_packet2 =
							    recv_info->
							    mux_packet;
							while (recv_packet2->
							       next) {
								recv_packet2 =
								    recv_packet2->
								    next;
							}
							recv_packet2->next =
							    recv_packet;
						}	/* End if( !(recv_info->mux_packet) ) */
					} else {	/* recv_info->total == 0 */
						if (uih_len >
						    TS0710MUX_RECV_BUF_SIZE) {
							TS0710_DEBUG
							    ("Error:  tty_idx:%d, uih_len == %d is too big\n",
							     tty_idx, uih_len);
							uih_len =
							    TS0710MUX_RECV_BUF_SIZE;
						}
						memcpy(recv_info->data,
						       uih_data_start, uih_len);
						recv_info->length = uih_len;
						recv_info->total = uih_len;

						add_post_recv_queue
						    (&mux_recv_queue,
						     recv_info);
					}	/* End recv_info->total == 0 */
				}	/* End Queue data */

				if (flow_control) {
					/* Do something for flow control */
					ts0710_flow_off(tty, dlci, ts0710);
				}

				if (post_recv) 
					schedule_work(&post_recv_tqueue);
			}	/* End processing received data */
		} else {
			TS0710_DEBUG("invalid dlci %d\n", dlci);
		}

		break;

	default:
		TS0710_DEBUG("illegal packet\n");
		break;
	}
	return 0;
}

/* Close ts0710 channel */
static void ts0710_close_channel(__u8 dlci)
{
	ts0710_con *ts0710 = &ts0710_connection;
	int try;
	unsigned long t;

	TS0710_DEBUG("ts0710_disc_command on channel %d\n", dlci);

	if ((ts0710->dlci[dlci].state == DISCONNECTED)
	    || (ts0710->dlci[dlci].state == REJECTED)) {
		return;
	} else if (ts0710->dlci[dlci].state == DISCONNECTING) {
		/* Reentry */
		return;
	} else {
		ts0710->dlci[dlci].state = DISCONNECTING;
		try = 3;
		while (try--) {
			t = jiffies;
			send_disc(ts0710, dlci);
			interruptible_sleep_on_timeout(&ts0710->dlci[dlci].
						       close_wait,
						       TS0710MUX_TIME_OUT);
			if (ts0710->dlci[dlci].state == DISCONNECTED) {
				break;
			} else if (signal_pending(current)) {
				TS0710_DEBUG
				    ("DLCI %d Send DISC got signal!\n",
				     dlci);
				break;
			} else if ((jiffies - t) >= TS0710MUX_TIME_OUT) {
				TS0710_DEBUG
				    ("DLCI %d Send DISC timeout!\n", dlci);
				continue;
			}
		}

		if (ts0710->dlci[dlci].state != DISCONNECTED) {
			if (dlci == 0) {	/* Control Channel */
				ts0710_upon_disconnect();
			} else {	/* Other Channel */
				ts0710->dlci[dlci].state = DISCONNECTED;
				wake_up_interruptible(&ts0710->dlci[dlci].
						      close_wait);
				ts0710_reset_dlci(dlci);
			}
		}
	}
}

static int ts0710_open_channel(__u8 dlci)
{
	ts0710_con *ts0710 = &ts0710_connection;
	int try;
	int retval;
	unsigned long t;

	retval = -ENODEV;
	if (dlci == 0) {	/* control channel */
		if ((ts0710->dlci[0].state == CONNECTED)
		    || (ts0710->dlci[0].state == FLOW_STOPPED)) {
			return 0;
		} else if (ts0710->dlci[0].state == CONNECTING) {
			/* Reentry */
			TS0710_DEBUG
			    ("DLCI: 0, re-entry to open DLCI 0, pid: %d, %s !\n",
			     current->pid, current->comm);
			try = 11;
			while (try--) {
				t = jiffies;
				interruptible_sleep_on_timeout(&ts0710->dlci[0].
							       open_wait,
							       TS0710MUX_TIME_OUT);
				if ((ts0710->dlci[0].state == CONNECTED)
				    || (ts0710->dlci[0].state ==
					FLOW_STOPPED)) {
					retval = 0;
					break;
				} else if (ts0710->dlci[0].state == REJECTED) {
					retval = -EREJECTED;
					break;
				} else if (ts0710->dlci[0].state ==
					   DISCONNECTED) {
					break;
				} else if (signal_pending(current)) {
					TS0710_DEBUG
					    ("DLCI:%d Wait for connecting got signal!\n",
					     dlci);
					retval = -EAGAIN;
					break;
				} else if ((jiffies - t) >= TS0710MUX_TIME_OUT) {
					TS0710_DEBUG
					    ("DLCI:%d Wait for connecting timeout!\n",
					     dlci);
					continue;
				} else if (ts0710->dlci[0].state == CONNECTING) {
					continue;
				}
			}

			if (ts0710->dlci[0].state == CONNECTING) {
				ts0710->dlci[0].state = DISCONNECTED;
			}
		} else if ((ts0710->dlci[0].state != DISCONNECTED)
			   && (ts0710->dlci[0].state != REJECTED)) {
			TS0710_DEBUG("DLCI:%d state is invalid!\n", dlci);
			return retval;
		} else {
			ts0710->initiator = 1;
			ts0710->dlci[0].state = CONNECTING;
			ts0710->dlci[0].initiator = 1;
			try = 10;
			while (try--) {
				t = jiffies;
				send_sabm(ts0710, 0);
				interruptible_sleep_on_timeout(&ts0710->dlci[0].
							       open_wait,
							       TS0710MUX_TIME_OUT);
				if ((ts0710->dlci[0].state == CONNECTED)
				    || (ts0710->dlci[0].state ==
					FLOW_STOPPED)) {
					retval = 0;
					break;
				} else if (ts0710->dlci[0].state == REJECTED) {
					TS0710_DEBUG
					    ("DLCI:%d Send SABM got rejected!\n",
					     dlci);
					retval = -EREJECTED;
					break;
				} else if (signal_pending(current)) {
					TS0710_DEBUG
					    ("DLCI:%d Send SABM got signal!\n",
					     dlci);
					retval = -EAGAIN;
					break;
				} else if ((jiffies - t) >= TS0710MUX_TIME_OUT) {
					TS0710_DEBUG
					    ("DLCI:%d Send SABM timeout!\n",
					     dlci);
					continue;
				}
			}

			if (ts0710->dlci[0].state == CONNECTING) {
				ts0710->dlci[0].state = DISCONNECTED;
			}
			wake_up_interruptible(&ts0710->dlci[0].open_wait);
		}
	} else {		/* other channel */
		if ((ts0710->dlci[0].state != CONNECTED)
		    && (ts0710->dlci[0].state != FLOW_STOPPED)) {
			return retval;
		} else if ((ts0710->dlci[dlci].state == CONNECTED)
			   || (ts0710->dlci[dlci].state == FLOW_STOPPED)) {
			return 0;
		} else if ((ts0710->dlci[dlci].state == NEGOTIATING)
			   || (ts0710->dlci[dlci].state == CONNECTING)) {
			/* Reentry */
			try = 8;
			while (try--) {
				t = jiffies;
				interruptible_sleep_on_timeout(&ts0710->
							       dlci[dlci].
							       open_wait,
							       TS0710MUX_TIME_OUT);
				if ((ts0710->dlci[dlci].state == CONNECTED)
				    || (ts0710->dlci[dlci].state ==
					FLOW_STOPPED)) {
					retval = 0;
					break;
				} else if (ts0710->dlci[dlci].state == REJECTED) {
					retval = -EREJECTED;
					break;
				} else if (ts0710->dlci[dlci].state ==
					   DISCONNECTED) {
					break;
				} else if (signal_pending(current)) {
					TS0710_DEBUG
					    ("DLCI:%d Wait for connecting got signal!\n",
					     dlci);
					retval = -EAGAIN;
					break;
				} else if ((jiffies - t) >= TS0710MUX_TIME_OUT) {
					TS0710_DEBUG
					    ("DLCI:%d Wait for connecting timeout!\n",
					     dlci);
					continue;
				} else
				    if ((ts0710->dlci[dlci].state ==
					 NEGOTIATING)
					|| (ts0710->dlci[dlci].state ==
					    CONNECTING)) {
					continue;
				}
			}

			if ((ts0710->dlci[dlci].state == NEGOTIATING)
			    || (ts0710->dlci[dlci].state == CONNECTING)) {
				ts0710->dlci[dlci].state = DISCONNECTED;
			}
		} else if ((ts0710->dlci[dlci].state != DISCONNECTED)
			   && (ts0710->dlci[dlci].state != REJECTED)) {
			TS0710_DEBUG("DLCI:%d state is invalid!\n", dlci);
			return retval;
		} else {
			ts0710->dlci[dlci].state = NEGOTIATING;
			ts0710->dlci[dlci].initiator = 1;
			try = 3;
			while (try--) {
				t = jiffies;
				send_pn_msg(ts0710, 7, ts0710->dlci[dlci].mtu,
					    0, 0, dlci, 1);
				interruptible_sleep_on_timeout(&ts0710->
							       dlci[dlci].
							       open_wait,
							       TS0710MUX_TIME_OUT);
				if (ts0710->dlci[dlci].state == CONNECTING) {
					break;
				} else if (signal_pending(current)) {
					TS0710_DEBUG
					    ("DLCI:%d Send pn_msg got signal!\n",
					     dlci);
					retval = -EAGAIN;
					break;
				} else if ((jiffies - t) >= TS0710MUX_TIME_OUT) {
					TS0710_DEBUG
					    ("DLCI:%d Send pn_msg timeout!\n",
					     dlci);
					continue;
				}
			}

			if (ts0710->dlci[dlci].state == CONNECTING) {
				try = 3;
				while (try--) {
					t = jiffies;
					send_sabm(ts0710, dlci);
					interruptible_sleep_on_timeout(&ts0710->
								       dlci
								       [dlci].
								       open_wait,
								       TS0710MUX_TIME_OUT);
					if ((ts0710->dlci[dlci].state ==
					     CONNECTED)
					    || (ts0710->dlci[dlci].state ==
						FLOW_STOPPED)) {
						retval = 0;
						break;
					} else if (ts0710->dlci[dlci].state ==
						   REJECTED) {
						TS0710_DEBUG
						    ("DLCI:%d Send SABM got rejected!\n",
						     dlci);
						retval = -EREJECTED;
						break;
					} else if (signal_pending(current)) {
						TS0710_DEBUG
						    ("DLCI:%d Send SABM got signal!\n",
						     dlci);
						retval = -EAGAIN;
						break;
					} else if ((jiffies - t) >=
						   TS0710MUX_TIME_OUT) {
						TS0710_DEBUG
						    ("DLCI:%d Send SABM timeout!\n",
						     dlci);
						continue;
					}
				}
			}

			if ((ts0710->dlci[dlci].state == NEGOTIATING)
			    || (ts0710->dlci[dlci].state == CONNECTING)) {
				ts0710->dlci[dlci].state = DISCONNECTED;
			}
			wake_up_interruptible(&ts0710->dlci[dlci].open_wait);
		}
	}
	return retval;
}

static int ts0710_exec_test_cmd(void)
{
	ts0710_con *ts0710 = &ts0710_connection;
	__u8 *f_buf;		/* Frame buffer */
	__u8 *d_buf;		/* Data buffer */
	int retval = -EFAULT;
	int j;
	unsigned long t;

	if (ts0710->be_testing) {
		/* Reentry */
		t = jiffies;
		interruptible_sleep_on_timeout(&ts0710->test_wait,
					       3 * TS0710MUX_TIME_OUT);
		if (ts0710->be_testing == 0) {
			if (ts0710->test_errs == 0) {
				retval = 0;
			} else {
				retval = -EFAULT;
			}
		} else if (signal_pending(current)) {
			TS0710_DEBUG
			    ("Wait for Test_cmd response got signal!\n");
			retval = -EAGAIN;
		} else if ((jiffies - t) >= 3 * TS0710MUX_TIME_OUT) {
			TS0710_DEBUG("Wait for Test_cmd response timeout!\n");
			retval = -EFAULT;
		}
	} else {
		ts0710->be_testing = 1;	/* Set the flag */

		f_buf = (__u8 *) kmalloc(TEST_PATTERN_SIZE + 32, GFP_KERNEL);
		d_buf = (__u8 *) kmalloc(TEST_PATTERN_SIZE + 32, GFP_KERNEL);
		if ((!f_buf) || (!d_buf)) {
			if (f_buf) {
				kfree(f_buf);
			}
			if (d_buf) {
				kfree(d_buf);
			}

			ts0710->be_testing = 0;	/* Clear the flag */
			ts0710->test_errs = TEST_PATTERN_SIZE;
			wake_up_interruptible(&ts0710->test_wait);
			return -ENOMEM;
		}

		for (j = 0; j < TEST_PATTERN_SIZE; j++) {
			d_buf[j] = j & 0xFF;
		}

		t = jiffies;
		ts0710_test_msg(ts0710, d_buf, TEST_PATTERN_SIZE, MCC_CMD,
				f_buf);
		interruptible_sleep_on_timeout(&ts0710->test_wait,
					       2 * TS0710MUX_TIME_OUT);
		if (ts0710->be_testing == 0) {
			if (ts0710->test_errs == 0) {
				retval = 0;
			} else {
				retval = -EFAULT;
			}
		} else if (signal_pending(current)) {
			TS0710_DEBUG("Send Test_cmd got signal!\n");
			retval = -EAGAIN;
		} else if ((jiffies - t) >= 2 * TS0710MUX_TIME_OUT) {
			TS0710_DEBUG("Send Test_cmd timeout!\n");
			ts0710->test_errs = TEST_PATTERN_SIZE;
			retval = -EFAULT;
		}

		ts0710->be_testing = 0;	/* Clear the flag */
		wake_up_interruptible(&ts0710->test_wait);

		/* Release buffer */
		if (f_buf) {
			kfree(f_buf);
		}
		if (d_buf) {
			kfree(d_buf);
		}
	}

	return retval;
}

static void mux_sched_send(void)
{
	if (!schedule_work(&send_tqueue)) {
		TS0710_DEBUG("mux_sched_send: Failed to schedule work\n");
	}
}

/****************************
 * TTY driver routines
*****************************/

static void mux_close(struct tty_struct *tty, struct file *filp)
{
	ts0710_con *ts0710 = &ts0710_connection;
	int line;
	__u8 dlci;

	UNUSED_PARAM(filp);

	if (!tty) {
		return;
	}
	line = tty->index;
	if ((line < TS0710MUX_MINOR_START) || (line >= NR_MUXS)) {
		TS0710_DEBUG("mux%d is not a valid port\n",line);
		return;
	}

	if (mux_tty[line] > 0)
		mux_tty[line]--;

	dlci = line;
	if (mux_tty[line] == 0)
		ts0710_close_channel(line);
	if (line == 1) {
		/* TODO - temporary, this assumes that dlc1 always exists and always last to close */
		TS0710_DEBUG("dlc 1 closing - therefore close cmd channel as well\n");

		/* TODO - for Siemens TC65 you need to send a CLD command - no further comms afterwards is possible */
		send_closedown_msg();
		ts0710_upon_disconnect();

		/* if not Siemens use ts0710_close_channel(0); */
        }

	TS0710_DEBUG("closing /dev/mux%d\n", line);

	if (mux_tty[line] == 0) {
		if ((mux_send_info_flags[line])
		    && (mux_send_info[line])
		    ) {
			mux_send_info_flags[line] = 0;
			kfree(mux_send_info[line]);
			mux_send_info[line] = 0;
		}

		if ((mux_recv_info_flags[line])
		    && (mux_recv_info[line])
		    && (mux_recv_info[line]->total == 0)) {
			mux_recv_info_flags[line] = 0;
			free_mux_recv_struct(mux_recv_info[line]);
			mux_recv_info[line] = 0;
		}

		ts0710_flow_on(dlci, ts0710);
		schedule_work(&post_recv_tqueue);

		wake_up_interruptible(&tty->read_wait);
		wake_up_interruptible(&tty->write_wait);
		tty->packet = 0;
	}
}

static void mux_throttle(struct tty_struct *tty)
{
	ts0710_con *ts0710 = &ts0710_connection;
	int line;
	int i;
	__u8 dlci;

	if (!tty) {
		return;
	}

	line = tty->index;
	if ((line < TS0710MUX_MINOR_START) || (line >= NR_MUXS)) {
		TS0710_DEBUG("mux%d is not a valid port\n",line);
		return;
	}

	TS0710_DEBUG("minor number is: %d\n", line);

	dlci = line;

	if ((ts0710->dlci[0].state != CONNECTED)
	    && (ts0710->dlci[0].state != FLOW_STOPPED)) {
		return;
	} else if ((ts0710->dlci[dlci].state != CONNECTED)
		   && (ts0710->dlci[dlci].state != FLOW_STOPPED)) {
		return;
	}

	if (ts0710->dlci[dlci].flow_control) {
		return;
	}

	for (i = 0; i < 3; i++) {
		if (ts0710_msc_msg
		    (ts0710, EA | FC | RTC | RTR | DV, MCC_CMD, dlci) < 0) {
			continue;
		} else {
			TS0710_LOG("Send Flow off on dlci %d\n", dlci);
			ts0710->dlci[dlci].flow_control = 1;
			break;
		}
	}
}

static void mux_unthrottle(struct tty_struct *tty)
{
	ts0710_con *ts0710 = &ts0710_connection;
	int line;
	__u8 dlci;
	mux_recv_struct *recv_info;

	if (!tty) {
		return;
	}
	line = tty->index;
	if ((line < TS0710MUX_MINOR_START) || (line >= NR_MUXS)) {
		TS0710_DEBUG("mux%d is not a valid port\n",line);
		return;
	}

	if ((!mux_recv_info_flags[line]) || (!mux_recv_info[line])) {
		return;
	}

	TS0710_DEBUG("minor number is: %d\n", line);

	recv_info = mux_recv_info[line];

	dlci = line;

	if (recv_info->total) {
		recv_info->post_unthrottle = 1;
		schedule_work(&post_recv_tqueue);
	} else {
		ts0710_flow_on(dlci, ts0710);
	}
}

static int mux_chars_in_buffer(struct tty_struct *tty)
{
	ts0710_con *ts0710 = &ts0710_connection;
	int retval;
	int line;
	__u8 dlci;
	mux_send_struct *send_info;

	retval = TS0710MUX_MAX_CHARS_IN_BUF;
	if (!tty) {
		goto out;
	}
	line = tty->index;
	if ((line < TS0710MUX_MINOR_START) || (line >= NR_MUXS)) {
		TS0710_DEBUG("mux%d is not a valid port\n",line);
		goto out;
	}

	dlci = line;

	if (ts0710->dlci[0].state == FLOW_STOPPED) {
		TS0710_DEBUG
		    ("Flow stopped on all channels, returning MAX chars in buffer\n");
		goto out;
	} else if (ts0710->dlci[dlci].state == FLOW_STOPPED) {
		TS0710_DEBUG("Flow stopped, returning MAX chars in buffer\n");
		goto out;
	} else if (ts0710->dlci[dlci].state != CONNECTED) {
		TS0710_DEBUG("DLCI %d not connected\n", dlci);
		goto out;
	}

	if (!(mux_send_info_flags[line])) {
		goto out;
	}
	send_info = mux_send_info[line];
	if (!send_info) {
		goto out;
	}
	if (send_info->filled) {
		goto out;
	}

	retval = 0;

      out:
	return retval;
}

static int mux_write(struct tty_struct *tty,
		     const unsigned char *buf, int count)
{
	ts0710_con *ts0710 = &ts0710_connection;
	int line;
	__u8 dlci;
	mux_send_struct *send_info;
	__u8 *d_buf;
	__u16 c;

	if (count <= 0) {
		return 0;
	}

	if (!tty) {
		return 0;
	}

	line = tty->index;
	if ((line < TS0710MUX_MINOR_START) || (line >= NR_MUXS)) {
		TS0710_DEBUG("mux%d is not a valid port\n",line);
		return -ENODEV;
        }

	dlci = line;

	if (ts0710->dlci[0].state == FLOW_STOPPED) {
		TS0710_DEBUG
		    ("Flow stopped on all channels, returning zero /dev/mux%d\n",
		     line);
		return 0;
	} else if (ts0710->dlci[dlci].state == FLOW_STOPPED) {
		TS0710_DEBUG("Flow stopped, returning zero /dev/mux%d\n", line);
		return 0;
	} else if (ts0710->dlci[dlci].state == CONNECTED) {

		if (!(mux_send_info_flags[line])) {
			TS0710_DEBUG
			    ("Error: mux_send_info_flags[%d] == 0\n",
			     line);
			return -ENODEV;
		}
		send_info = mux_send_info[line];
		if (!send_info) {
			TS0710_DEBUG
			    ("Error: mux_send_info[%d] == 0\n",
			     line);
			return -ENODEV;
		}

		c = min(count, (ts0710->dlci[dlci].mtu - 1));
		if (c <= 0) {
			return 0;
		}

		if (test_and_set_bit(BUF_BUSY, &send_info->flags))
			return 0;

		if (send_info->filled) {
			clear_bit(BUF_BUSY, &send_info->flags);
			return 0;
		}

		d_buf = ((__u8 *) send_info->buf) + TS0710MUX_SEND_BUF_OFFSET;

		memcpy(d_buf, buf, c);

		TS0710_DEBUGHEX(d_buf, c + 1);

		send_info->frame = d_buf;
		queue_uih(send_info, c + 1, ts0710, dlci);
		send_info->filled = 1;
		clear_bit(BUF_BUSY, &send_info->flags);

		if (mux_ldisc_chars_in_buffer(COMM_FOR_MUX_TTY) == 0) {
			/* Sending bottom half should be
			   run after return from this function */
			mux_sched_send();
		}
		return c;
	} else {
		TS0710_DEBUG("DLCI %d not connected\n", dlci);
		return -EDISCONNECTED;
	}
}

static int mux_write_room(struct tty_struct *tty)
{
	ts0710_con *ts0710 = &ts0710_connection;
	int retval;
	int line;
	__u8 dlci;
	mux_send_struct *send_info;

	retval = 0;
	if (!tty) {
		goto out;
	}
	line = tty->index;
	if ((line < TS0710MUX_MINOR_START) || (line >= NR_MUXS)) {
		TS0710_DEBUG("mux%d is not a valid port\n",line);
		goto out;
	}

	dlci = line;

	if (ts0710->dlci[0].state == FLOW_STOPPED) {
		TS0710_DEBUG("Flow stopped on all channels, returning ZERO\n");
		goto out;
	} else if (ts0710->dlci[dlci].state == FLOW_STOPPED) {
		TS0710_DEBUG("Flow stopped, returning ZERO\n");
		goto out;
	} else if (ts0710->dlci[dlci].state != CONNECTED) {
		TS0710_DEBUG("DLCI %d not connected\n", dlci);
		goto out;
	}

	if (!(mux_send_info_flags[line])) {
		goto out;
	}
	send_info = mux_send_info[line];
	if (!send_info) {
		goto out;
	}
	if (send_info->filled) {
		goto out;
	}

	retval = ts0710->dlci[dlci].mtu - 1;

      out:
	return retval;
}

static int mux_ioctl(struct tty_struct *tty, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	ts0710_con *ts0710 = &ts0710_connection;
	int line;
	__u8 dlci;

	UNUSED_PARAM(file);
	UNUSED_PARAM(arg);

	if (!tty) {
		return -EIO;
	}
	line = tty->index;
	if ((line < TS0710MUX_MINOR_START) || (line >= NR_MUXS)) {
		TS0710_DEBUG("mux%d is not a valid port\n",line);
		return -ENODEV;
	}

	dlci = line;

	switch (cmd) {
	case TS0710MUX_IO_MSC_HANGUP:
		if (ts0710_msc_msg(ts0710, EA | RTR | DV, MCC_CMD, dlci) < 0) {
			return -EAGAIN;
		} else {
			return 0;
		}

	case TS0710MUX_IO_TEST_CMD:
		return ts0710_exec_test_cmd();
	default:
		break;
	}
	return -ENOIOCTLCMD;
}

static void mux_flush_buffer(struct tty_struct *tty)
{
	int line;

	if (!tty) {
		return;
	}

	line = tty->index;
	if ((line < TS0710MUX_MINOR_START) || (line >= NR_MUXS)) {
		TS0710_DEBUG("mux%d is not a valid port\n",line);
		return;
	}

	if ((mux_send_info_flags[line])
	    && (mux_send_info[line])
	    && (mux_send_info[line]->filled)) {

		mux_send_info[line]->filled = 0;
	}

	wake_up_interruptible(&tty->write_wait);
#ifdef SERIAL_HAVE_POLL_WAIT
	wake_up_interruptible(&tty->poll_wait);
#endif
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	    tty->ldisc.write_wakeup) {
		(tty->ldisc.write_wakeup) (tty);
	}
}

static int mux_open(struct tty_struct *tty, struct file *filp)
{
	int retval = -ENODEV;
	int line = 0;
	__u8 dlci;

	mux_send_struct *send_info;
	mux_recv_struct *recv_info;

	UNUSED_PARAM(filp);

	if (!tty) {
		goto out;
	}
	line = tty->index;
	if ((line < TS0710MUX_MINOR_START) || (line >= NR_MUXS)) {
		TS0710_DEBUG("mux%d is not a valid port\n",line);
		goto out;
	}

	mux_tty[line]++;

	dlci = line;

	mux_table[line] = tty;

	/* Open server channel 0 first */
	if ((retval = ts0710_open_channel(0)) != 0) {
		TS0710_DEBUG("Can't connect server channel 0!\n");
		ts0710_init();

		mux_tty[line]--;
		goto out;
	}

	/* Allocate memory first. As soon as connection has been established, MUX may receive */
	if (mux_send_info_flags[line] == 0) {
		send_info =
		    (mux_send_struct *) kmalloc(sizeof(mux_send_struct),
						GFP_KERNEL);
		if (!send_info) {
			retval = -ENOMEM;

			mux_tty[line]--;
			goto out;
		}
		send_info->length = 0;
		send_info->flags = 0;
		send_info->filled = 0;
		mux_send_info[line] = send_info;
		mux_send_info_flags[line] = 1;
	}

	if (mux_recv_info_flags[line] == 0) {
		recv_info =
		    (mux_recv_struct *) kmalloc(sizeof(mux_recv_struct),
						GFP_KERNEL);
		if (!recv_info) {
			mux_send_info_flags[line] = 0;
			kfree(mux_send_info[line]);
			mux_send_info[line] = 0;
			retval = -ENOMEM;

			mux_tty[line]--;
			goto out;
		}
		recv_info->length = 0;
		recv_info->total = 0;
		recv_info->mux_packet = 0;
		recv_info->next = 0;
		recv_info->no_tty = line;
		recv_info->post_unthrottle = 0;
		mux_recv_info[line] = recv_info;
		mux_recv_info_flags[line] = 1;
	}

	/* Now establish DLCI connection */
	if ((retval = ts0710_open_channel(dlci)) != 0) {
		TS0710_DEBUG("Can't connected channel %d!\n",dlci);
		ts0710_reset_dlci(dlci);

		mux_send_info_flags[line] = 0;
		kfree(mux_send_info[line]);
		mux_send_info[line] = 0;

		mux_recv_info_flags[line] = 0;
		free_mux_recv_struct(mux_recv_info[line]);
		mux_recv_info[line] = 0;

		mux_tty[line]--;
		goto out;
	}

	retval = 0;
        TS0710_DEBUG("opening /dev/mux%d\n", line);

      out:
	TS0710_DEBUG("returning %d (mux_tty[%d] = %d)\n", retval, line, mux_tty[line]);
	return retval;
}

static void mux_dispatcher(struct tty_struct *tty)
{
	UNUSED_PARAM(tty);
	schedule_work(&receive_tqueue);
}

static unsigned char* processed_received_data(unsigned char *start, unsigned char* end)
{
	int count;
	static unsigned char *start_flag = 0;
	unsigned char *to, *from;
	short_frame *short_pkt;
	long_frame *long_pkt;
	static int framelen = -1;

        unsigned char *search = start;
        unsigned char *tbuf_ptr = end;

	__u32 crc_error;
	__u8 *uih_data_start;
	__u32 uih_len;

	while (1) {
		if (start_flag == 0) {	/* Frame Start Flag not found */
			framelen = -1;
			while (search < tbuf_ptr) {
				if (*search == TS0710_BASIC_FLAG) {
					start_flag = search;
					break;
				}

				search++;
			}

			if (start_flag == 0) {
				tbuf_ptr = start;
				break;
			}
		} else {	/* Frame Start Flag found */
			/* 1 start flag + 1 address + 1 control + 1 or 2 length + lengths data + 1 FCS + 1 end flag */
			/* For BP UART problem 1 start flag + 1 seq_num + 1 address + ...... */
			/*if( (framelen == -1) && ((tbuf_ptr - start_flag) > TS0710_MAX_HDR_SIZE) ) */
			if ((framelen == -1) && ((tbuf_ptr - start_flag) > (TS0710_MAX_HDR_SIZE + SEQ_FIELD_SIZE))) {	/*For BP UART problem */
				/*short_pkt = (short_frame *) (start_flag + 1); */
				short_pkt = (short_frame *) (start_flag + ADDRESS_FIELD_OFFSET);	/*For BP UART problem */
				if (short_pkt->h.length.ea == 1) {	/* short frame */
					/*framelen = TS0710_MAX_HDR_SIZE + short_pkt->h.length.len + 1; */
					framelen = TS0710_MAX_HDR_SIZE + short_pkt->h.length.len + 1 + SEQ_FIELD_SIZE;	/*For BP UART problem */
				} else {	/* long frame */
					/*long_pkt = (long_frame *) (start_flag + 1); */
					long_pkt = (long_frame *) (start_flag + ADDRESS_FIELD_OFFSET);	/*For BP UART problem */
					/*framelen = TS0710_MAX_HDR_SIZE + GET_LONG_LENGTH( long_pkt->h.length ) + 2; */
					framelen = TS0710_MAX_HDR_SIZE + GET_LONG_LENGTH(long_pkt->h.length) + 2 + SEQ_FIELD_SIZE;	/*For BP UART problem */
				}

				/*if( framelen > TS0710MUX_MAX_TOTAL_FRAME_SIZE ) { */
				if (framelen > (TS0710MUX_MAX_TOTAL_FRAME_SIZE + SEQ_FIELD_SIZE)) {	/*For BP UART problem */
					TS0710_LOGSTR_FRAME(0, start_flag,
							    (tbuf_ptr -
							     start_flag));
					TS0710_DEBUG
					    ("Error: frame length:%d is bigger than Max total frame size:%d\n",
					     framelen, (TS0710MUX_MAX_TOTAL_FRAME_SIZE + SEQ_FIELD_SIZE));	/*For BP UART problem */
					search = start_flag + 1;
					start_flag = 0;
					framelen = -1;
					continue;
				}
			}

			if ((framelen != -1)
			    && ((tbuf_ptr - start_flag) >= framelen)) {
				if (*(start_flag + framelen - 1) == TS0710_BASIC_FLAG) {	/* OK, We got one frame */

					TS0710_LOGSTR_FRAME(0, start_flag,
							    framelen);
					TS0710_DEBUGHEX(start_flag, framelen);

					short_pkt =
					    (short_frame *) (start_flag +
							     ADDRESS_FIELD_OFFSET);
					if ((short_pkt->h.length.ea) == 0) {
						long_pkt =
						    (long_frame *) (start_flag +
								    ADDRESS_FIELD_OFFSET);
						uih_len =
						    GET_LONG_LENGTH(long_pkt->h.
								    length);
						uih_data_start =
						    long_pkt->h.data;

						crc_error =
						    crc_check((__u8
							       *) (start_flag +
								   SLIDE_BP_SEQ_OFFSET),
							      LONG_CRC_CHECK
							      ,*(uih_data_start +
								uih_len));
					} else {
						uih_len =
						    short_pkt->h.length.len;
						uih_data_start =
						    short_pkt->data;

						crc_error =
						    crc_check((__u8
							       *) (start_flag +
								   SLIDE_BP_SEQ_OFFSET),
							      SHORT_CRC_CHECK
							      ,*(uih_data_start +
								uih_len));
					}

					if (!crc_error) {
							ts0710_recv_data
							    (&ts0710_connection,
							     start_flag +
							     ADDRESS_FIELD_OFFSET,
							     framelen - 2 -
							     SEQ_FIELD_SIZE);
					} else {	/* crc_error */
						search = start_flag + 1;
						start_flag = 0;
						framelen = -1;
						continue;
					}	/*End if(!crc_error) */

					search = start_flag + framelen;
				} else {
					TS0710_LOGSTR_FRAME(0, start_flag,
							    framelen);
					TS0710_DEBUGHEX(start_flag, framelen);
					TS0710_DEBUG("Lost synchronization!\n");
					search = start_flag + 1;
				}

				start_flag = 0;
				framelen = -1;
				continue;
			}

			if (start_flag != start) {
				to = start;
				from = start_flag;
				count = tbuf_ptr - start_flag;
				while (count--) {
					*to++ = *from++;
				}

				tbuf_ptr -= (start_flag - start);
				start_flag = start;
			}
			break;
		}		/* End Frame Start Flag found */
	}			/* End while(1) */
        return tbuf_ptr;
}

static unsigned char tbuf[TS0710MUX_MAX_BUF_SIZE];
static unsigned char *tbuf_ptr = tbuf;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static void receive_worker(void *private_)
#else
static void receive_worker(struct work_struct *private_)
#endif
{
	int count, tbuf_free, tbuf_read;

	UNUSED_PARAM(private_);

	if (!COMM_FOR_MUX_TTY)
		return;
 
	if (test_and_set_bit(RECV_RUNNING, &mux_recv_flags)) {
		schedule_work(&receive_tqueue);
		return;
	}

	TS0710_DEBUG(">>entry\n");

	while (1) {
		tbuf_free = TS0710MUX_MAX_BUF_SIZE - (tbuf_ptr - tbuf);
		TS0710_DEBUG("%i bytes free\n", tbuf_free);
		tbuf_read = get_from_inbuf_list(tbuf_ptr, tbuf_free);
		if (tbuf_read == 0) {
			TS0710_DEBUG("finished read\n");
			break;
		} else {
			TS0710_DEBUG("read %i bytes.\n", tbuf_read);
		};
		tbuf_ptr += tbuf_read;
	};
 
	count = (tbuf_ptr - tbuf);

	/* received the following bytes */
	TS0710_DEBUGHEX(tbuf, count);
 
        tbuf_ptr = processed_received_data(tbuf,tbuf_ptr);

	clear_bit(RECV_RUNNING, &mux_recv_flags);
	TS0710_DEBUG("<<exit\n");
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static void post_recv_worker(void *private_)
#else
static void post_recv_worker(struct work_struct *private_)
#endif
{
	ts0710_con *ts0710 = &ts0710_connection;
	int tty_idx;
	struct tty_struct *tty;
	__u8 post_recv;
	__u8 flow_control;
	__u8 dlci;
	mux_recv_struct *recv_info, *recv_info2, *post_recv_q;
	int recv_room;
	mux_recv_packet *recv_packet, *recv_packet2;

	UNUSED_PARAM(private_);

	if (test_and_set_bit(RECV_RUNNING, &mux_recv_flags)) {
		schedule_work(&post_recv_tqueue);
		return;
	}

	TS0710_DEBUG(">>entry mux_recv_queue %d\n",(int) mux_recv_queue);

	post_recv = 0;
	if (!mux_recv_queue) {
		goto out;
	}

	post_recv_q = NULL;
	recv_info2 = mux_recv_queue;
	while ((recv_info = recv_info2)) {
		recv_info2 = recv_info->next;

		if (!(recv_info->total)) {
			TS0710_DEBUG
			    ("Error: Should not get here, recv_info->total == 0 \n");
			continue;
		}

		tty_idx = recv_info->no_tty;

		dlci = tty_idx;

		tty = mux_table[tty_idx];
		if ((!mux_tty[tty_idx]) || (!tty)) {
			TS0710_DEBUG
			    ("No application waiting for, free recv_info! tty_idx:%d\n",
			     tty_idx);
			mux_recv_info_flags[tty_idx] = 0;
			free_mux_recv_struct(mux_recv_info[tty_idx]);
			mux_recv_info[tty_idx] = 0;
			ts0710_flow_on(dlci, ts0710);
			continue;
		}

		TS0710_DEBUG("/dev/mux%d recv_info->total is: %d\n", tty_idx,
			     recv_info->total);

		if (test_bit(TTY_THROTTLED, &tty->flags)) {
			add_post_recv_queue(&post_recv_q, recv_info);
			continue;
		}

		flow_control = 0;
		recv_packet2 = recv_info->mux_packet;
		while (recv_info->total) {
			recv_room = 65535;
			if (tty->receive_room)
				recv_room = tty->receive_room;

			if (recv_info->length) {
				if (recv_room < recv_info->length) {
					flow_control = 1;
					break;
				}

				/* Put queued data into read buffer of tty */
				TS0710_DEBUG
				    ("Put queued recv data into read buffer of /dev/mux%d\n",
				     tty_idx);
				TS0710_DEBUGHEX(recv_info->data,
						recv_info->length);
				(tty->ldisc.receive_buf) (tty, recv_info->data,
							  NULL,
							  recv_info->length);
				recv_info->total -= recv_info->length;
				recv_info->length = 0;
			} else {	/* recv_info->length == 0 */
				if ((recv_packet = recv_packet2)) {
					recv_packet2 = recv_packet->next;

					if (recv_room < recv_packet->length) {
						flow_control = 1;
						recv_info->mux_packet =
						    recv_packet;
						break;
					}

					/* Put queued data into read buffer of tty */
					TS0710_DEBUG
					    ("Put queued recv data into read buffer of /dev/mux%d\n",
					     tty_idx);
					TS0710_DEBUGHEX(recv_packet->data,
							recv_packet->length);
					(tty->ldisc.receive_buf) (tty,
								  recv_packet->
								  data, NULL,
								  recv_packet->
								  length);
					recv_info->total -= recv_packet->length;
					free_mux_recv_packet(recv_packet);
				} else {
					TS0710_DEBUG
					    ("Error: Should not get here, recv_info->total is:%u \n", recv_info->total);
				}
			}	/* End recv_info->length == 0 */
		}		/* End while( recv_info->total ) */

		if (!(recv_info->total)) {
			/* Important clear */
			recv_info->mux_packet = 0;

			if (recv_info->post_unthrottle) {
				/* Do something for post_unthrottle */
				ts0710_flow_on(dlci, ts0710);
				recv_info->post_unthrottle = 0;
			}
		} else {
			add_post_recv_queue(&post_recv_q, recv_info);

			if (flow_control) {
				/* Do something for flow control */
				if (recv_info->post_unthrottle) {
					set_bit(TTY_THROTTLED, &tty->flags);
					recv_info->post_unthrottle = 0;
				} else {
					ts0710_flow_off(tty, dlci, ts0710);
				}
			}	/* End if( flow_control ) */
		}
	}			/* End while( (recv_info = recv_info2) ) */

	mux_recv_queue = post_recv_q;

      out:
	clear_bit(RECV_RUNNING, &mux_recv_flags);
	TS0710_DEBUG("<<exit\n");
}

static void mux_sender(void)
{
	mux_send_struct *send_info;
	int chars;
	__u8 idx;

	chars = mux_ldisc_chars_in_buffer(COMM_FOR_MUX_TTY);
	if (!chars) {
		TS0710_DEBUG("no chars to send\n");
		mux_sched_send();
		return;
	}

	idx = mux_send_info_idx;
	if ((idx < NR_MUXS) && (mux_send_info_flags[idx])) {
		send_info = mux_send_info[idx];
		if ((send_info)
		    && (send_info->filled)
		    && (send_info->length <=
			(TS0710MUX_SERIAL_BUF_SIZE - chars))) {

			mux_sched_send();
		}
	}
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static void send_worker(void *private_)
#else
static void send_worker(struct work_struct *private_)
#endif
{
	ts0710_con *ts0710 = &ts0710_connection;
	__u8 j;
	mux_send_struct *send_info;
	int chars;
	struct tty_struct *tty;
	__u8 dlci;

	UNUSED_PARAM(private_);

	TS0710_DEBUG(">>entry\n");

	mux_send_info_idx = NR_MUXS;

	if (ts0710->dlci[0].state == FLOW_STOPPED) {
		TS0710_DEBUG("Flow stopped on all channels\n");
		return;
	}

	for (j = 0; j < NR_MUXS; j++) {

		if (!(mux_send_info_flags[j])) {
			continue;
		}

		send_info = mux_send_info[j];
		if (!send_info) {
			continue;
		}

		if (!(send_info->filled)) {
			continue;
		}

		dlci = j;

		if (ts0710->dlci[dlci].state == FLOW_STOPPED) {
			TS0710_DEBUG("Flow stopped on channel DLCI: %d\n",
				     dlci);
			continue;
		} else if (ts0710->dlci[dlci].state != CONNECTED) {
			TS0710_DEBUG("DLCI %d not connected\n", dlci);
			send_info->filled = 0;
			continue;
		}

		chars = mux_ldisc_chars_in_buffer(COMM_FOR_MUX_TTY);
		if (send_info->length <= (TS0710MUX_SERIAL_BUF_SIZE - chars)) {
			TS0710_DEBUG("Send queued UIH for /dev/mux%d\n", j);
			basic_write((__u8 *) send_info->frame,send_info->length);
			send_info->length = 0;
			send_info->filled = 0;
		} else {
			mux_send_info_idx = j;
			break;
		}
	}			/* End for() loop */

	/* Queue UIH data to be transmitted */
	for (j = 0; j < NR_MUXS; j++) {

		if (!(mux_send_info_flags[j])) {
			continue;
		}

		send_info = mux_send_info[j];
		if (!send_info) {
			continue;
		}

		if (send_info->filled) {
			continue;
		}

		/* Now queue UIH data to send_info->buf */

		if (!mux_tty[j]) {
			continue;
		}

		tty = mux_table[j];
		if (!tty) {
			continue;
		}

		dlci = j;

		if (ts0710->dlci[dlci].state == FLOW_STOPPED) {
			TS0710_DEBUG("Flow stopped on channel DLCI: %d\n",
				     dlci);
			continue;
		} else if (ts0710->dlci[dlci].state != CONNECTED) {
			TS0710_DEBUG("DLCI %d not connected\n", dlci);
			continue;
		}

		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP))
		    && tty->ldisc.write_wakeup) {
			(tty->ldisc.write_wakeup) (tty);
		}
		wake_up_interruptible(&tty->write_wait);

#ifdef SERIAL_HAVE_POLL_WAIT
		wake_up_interruptible(&tty->poll_wait);
#endif

		if (send_info->filled) {
			if (j < mux_send_info_idx) {
				mux_send_info_idx = j;
			}
		}
	}			/* End for() loop */
	TS0710_DEBUG("<<exit\n");
}

/*********************************************************************************************/
/* line discipline */

/* Taken from ts0710_mux_usb.c */
static void append_to_inbuf_list(const unsigned char *buf, int count)
{
	buf_list_t *inbuf;
	
	inbuf = kmalloc(sizeof(buf_list_t), GFP_KERNEL);
	if (!inbuf) {
		TS0710_DEBUG("(%d) out of memory!\n",sizeof(buf_list_t));
		return;
	}
	
	inbuf->size = count;

	inbuf->body = kmalloc(sizeof(char)*count, GFP_KERNEL);
	if (!inbuf->body) {
		kfree(inbuf);
		TS0710_DEBUG("(%d) out of memory!\n",sizeof(unsigned char)*count);
		return;
	}
	memcpy(inbuf->body, buf, count);
	spin_lock(&ldisc_data_ptr->in_buf_lock);
        ldisc_data_ptr->in_buf_size += count;
	list_add_tail(&inbuf->list, &ldisc_data_ptr->in_buf_list);
	spin_unlock(&ldisc_data_ptr->in_buf_lock);
}

/* Taken from ts0710_mux_usb.c */
static int get_from_inbuf_list(unsigned char *buf, int dst_count)
{
	int ret = 0;
	spin_lock(&ldisc_data_ptr->in_buf_lock);
	if (!(list_empty(&ldisc_data_ptr->in_buf_list))) { /* only removes one */
		int src_count;
		buf_list_t *inbuf;
		struct list_head *ptr;

		ptr = ldisc_data_ptr->in_buf_list.next;
		inbuf = list_entry(ptr, buf_list_t, list);
		src_count = inbuf->size;
		if (dst_count >= src_count) {
			memcpy(buf, inbuf->body, src_count);
			ret = src_count;
                        ldisc_data_ptr->in_buf_size -= src_count;
			list_del(ptr);
			kfree(inbuf->body);
			kfree(inbuf);
		} else {
	               TS0710_DEBUG("not enough space in destination buffer\n");
		}
	}
	spin_unlock(&ldisc_data_ptr->in_buf_lock);

	return ret;
}

static int mux_ldisc_write(const unsigned char *buf, int count)
{	
	int ret = COMM_FOR_MUX_TTY->driver->write(COMM_FOR_MUX_TTY, buf, count);

        mux_sender();

        return ret;
}

static int mux_ldisc_chars_in_buffer(struct tty_struct *tty)
{		
	TS0710_DEBUG("%d\n",ldisc_data_ptr->in_buf_size);
	return ldisc_data_ptr->in_buf_size;
}

static int
link_ldisc_open(struct tty_struct *tty)
{
	if (!tty)
		return -EINVAL;

	TS0710_DEBUG("link_ldisc_open\n");
        /* TODO - force connection of the mux - assumption that this is called after cmux issued */

	COMM_FOR_MUX_TTY = tty;
	tty->disc_data = NULL;
	tty->receive_room = 65536;

	return 0;
}

static void
link_ldisc_close(struct tty_struct *tty)
{
	TS0710_DEBUG("link_ldisc_close\n\n");

        /* TODO - force disconnection of the mux */
}

static int link_ldisc_hangup(struct tty_struct *tty)
{
	TS0710_DEBUG("link_ldisc_hangup\n");
	link_ldisc_close(tty);
	return 0;
}

static ssize_t
link_ldisc_read(struct tty_struct *tty, struct file *file,
		  unsigned char __user *buf, size_t count)
{
        /* TODO - we could give back diagnostic info in this read since /dev/muxX used for data reads */
	TS0710_DEBUG("link_ldisc_read\n");
	return -EAGAIN;
}

static ssize_t
link_ldisc_write(struct tty_struct *tty, struct file *file,
		   const unsigned char *buf, size_t count)
{
         /* TODO - write currently not used - /dev/muxX used for data writes */
       TS0710_DEBUG("link_ldisc_write - do nothing with <%s>\n",buf);
        return 0;
}

static int
link_ldisc_ioctl(struct tty_struct *tty, struct file *file,
		   unsigned int cmd, unsigned long arg)
{
	TS0710_DEBUG("link_ldisc_ioctl\n");
        /* TODO - any ioctls need processing */
	return 0;
}

static unsigned int
link_ldisc_poll(struct tty_struct *tty, struct file *file, poll_table *wait)
{
	TS0710_DEBUG("link_ldisc_poll\n");
        /* TODO - see link_ldisc_read comment above - only poll for diagnostic info */
	return 0;
}

static void
link_ldisc_receive(struct tty_struct *tty, const unsigned char *buf,
		  char *cflags, int count)
{
	TS0710_DEBUG(">>entry count %d\n",count);

	/* called by tty_io's flush_to_ldisc method */
	/* if (tty->low_latency) */
	{
		/* user has specified low latency but we cannot do this quickly
		   so store for later & use work queue */
		{
			/* Add to inbuf so can later be retrieved by get_from_inbuf_list */
			append_to_inbuf_list(buf,count);

			/* schedule the receive work queue */
			mux_dispatcher(tty);
		}
	}
#if 0
/* TODO - following code needs checking but likely to be marginally faster when working correctly */
	else
	{
		/* This is being called in normal content by tty_io buf.work work
		   queue so we can process the data now */
		int tbuf_free = TS0710MUX_MAX_BUF_SIZE - (tbuf_ptr - tbuf);
		TS0710_DEBUG("%i bytes free\n", tbuf_free);
		if (tbuf_free > 0) {
			int copy_size = count;
			if (tbuf_free < count) {
				/* TODO - losing data */
				TS0710_DEBUG("Some data loss!\n");
				copy_size = tbuf_free;
			}
			memcpy(tbuf_ptr, buf, copy_size);
			tbuf_ptr = processed_received_data(tbuf_ptr,(tbuf_ptr + copy_size));
		} else {
			/* TODO - losing data */
			TS0710_DEBUG("Run out of buffer space!\n");
		}
	}
#endif

	TS0710_DEBUG("<<exit\n");
}

static void
link_ldisc_wakeup(struct tty_struct *tty)
{
	TS0710_DEBUG("link_ldisc_wakeup\n");
	/* TODO - any use for wakeup? */
}


static struct tty_ldisc tty_ldisc_info = {
	.owner  = THIS_MODULE,
	.magic	= TTY_LDISC_MAGIC,
	.name	= "link_ldisc",
	.open	= link_ldisc_open,
	.close	= link_ldisc_close,
	.hangup	= link_ldisc_hangup,
	.read	        = link_ldisc_read,
	.write	= link_ldisc_write,
	.ioctl	        = link_ldisc_ioctl,
	.poll	        = link_ldisc_poll,
	.receive_buf = link_ldisc_receive,
	.write_wakeup = link_ldisc_wakeup,
};

static int link_ldisc_init(void)
{
	int err;
	TS0710_DEBUG("link_ldisc_init\n");

	if (!(ldisc_data_ptr = kzalloc(sizeof(struct link_ldisc_data), GFP_KERNEL))) {
		TS0710_DEBUG("link_ldisc_init: Out of memory.\n");		
		return -ENOMEM;
	}
	INIT_LIST_HEAD(&ldisc_data_ptr->in_buf_list);
	ldisc_data_ptr->in_buf_lock = SPIN_LOCK_UNLOCKED;

	/* Reuse the N_MOUSE ldisc for now, TODO - need own line discipline ident */
	err = tty_register_ldisc(N_MOUSE, &tty_ldisc_info);
	if (err != 0) {
		printk(KERN_ERR "link_ldisc: error %d registering line disc.\n",err);
	}

	return err;
}

static void link_ldisc_cleanup(void)
{
	TS0710_DEBUG("link_ldisc_cleanup\n");

	if (tty_unregister_ldisc(N_MOUSE) != 0) {
		printk(KERN_ERR "failed to unregister line discipline\n");
	}
	kfree(ldisc_data_ptr);
}

/*********************************************************************************************/
/* Module init & exit */

static int __init mux_init(void)
{
	__u8 j;

	TS0710_DEBUG("mux_init\n");
	TS0710_DEBUG("The process is \"%s\" (pid %i)\n", current->comm, current->pid);

        /* set up the line discipline first */
	if (link_ldisc_init()) {
	       panic("Failed to initialise the line discipline\n");
	}

	ts0710_init();

	for (j = 0; j < NR_MUXS; j++) {
		mux_send_info_flags[j] = 0;
		mux_send_info[j] = 0;
		mux_recv_info_flags[j] = 0;
		mux_recv_info[j] = 0;
	}
	mux_send_info_idx = NR_MUXS;
	mux_recv_queue = NULL;
	mux_recv_flags = 0;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
	INIT_WORK(&send_tqueue, send_worker, NULL);
	INIT_WORK(&receive_tqueue, receive_worker, NULL);
	INIT_WORK(&post_recv_tqueue, post_recv_worker, NULL);
#else
	INIT_WORK(&send_tqueue, send_worker);
	INIT_WORK(&receive_tqueue, receive_worker);
	INIT_WORK(&post_recv_tqueue, post_recv_worker);
#endif

	mux_driver = alloc_tty_driver(NR_MUXS);
	if (!mux_driver)
		return -ENOMEM;

	mux_driver->owner = THIS_MODULE;
	mux_driver->driver_name = "ts0710mux";
	mux_driver->name = "mux";
	mux_driver->major = TS0710MUX_MAJOR;
	mux_driver->minor_start = TS0710MUX_MINOR_START;
	mux_driver->type = TTY_DRIVER_TYPE_SERIAL;
	mux_driver->subtype = SERIAL_TYPE_NORMAL;
	mux_driver->flags = TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW;

	mux_driver->init_termios = tty_std_termios;
	mux_driver->init_termios.c_iflag = 0;
	mux_driver->init_termios.c_oflag = 0;
	mux_driver->init_termios.c_cflag = B38400 | CS8 | CREAD;
	mux_driver->init_termios.c_lflag = 0;

	mux_driver->termios = mux_termios;
	mux_driver->termios_locked = mux_termios_locked;
	mux_driver->other = NULL;

	mux_driver->open = mux_open;
	mux_driver->close = mux_close;
	mux_driver->write = mux_write;
	mux_driver->write_room = mux_write_room;
	mux_driver->flush_buffer = mux_flush_buffer;
	mux_driver->chars_in_buffer = mux_chars_in_buffer;
	mux_driver->throttle = mux_throttle;
	mux_driver->unthrottle = mux_unthrottle;
	mux_driver->ioctl = mux_ioctl;

	/* FIXME: No panic() here */
	if (tty_register_driver(mux_driver))
		panic("Couldn't register mux driver");

	for (j=TS0710MUX_MINOR_START; j<NR_MUXS; j++)
		tty_register_device(mux_driver, j, NULL);

	return 0;
}

static void __exit mux_exit(void)
{
	__u8 j;

	mux_send_info_idx = NR_MUXS;
	mux_recv_queue = NULL;
	for (j = 0; j < NR_MUXS; j++) {
		if ((mux_send_info_flags[j]) && (mux_send_info[j])) {
			kfree(mux_send_info[j]);
		}
		mux_send_info_flags[j] = 0;
		mux_send_info[j] = 0;

		if ((mux_recv_info_flags[j]) && (mux_recv_info[j])) {
			free_mux_recv_struct(mux_recv_info[j]);
		}
		mux_recv_info_flags[j] = 0;
		mux_recv_info[j] = 0;
	}

	for (j=TS0710MUX_MINOR_START; j<NR_MUXS; j++)
		tty_unregister_device(mux_driver, j);

	if (tty_unregister_driver(mux_driver))
		panic("Couldn't unregister mux driver");

	/* remove the line discipline last */
	link_ldisc_cleanup();
}

module_init(mux_init);
module_exit(mux_exit);

/****************************************************************************************************************/

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jim Rayner <jimr@beyondvoice.com>");
MODULE_DESCRIPTION("27.010 multiplexer");
